# -*- coding: utf-8 -*-

"""
This file contains modified pyqtgraph.ImageItem subclasses for data visualization.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['DataImageItem', 'XYPlotItem']

import numpy as np
from typing import Union, Optional, Tuple
from PySide2 import QtCore
from pyqtgraph import ImageItem as _ImageItem
from pyqtgraph import PlotDataItem as _PlotDataItem

from qudi.util.colordefs import ColorScaleInferno as _Colorscale
from qudi.util.colordefs import QudiPalette as _QudiPalette


class XYPlotItem(_PlotDataItem):
    """ Extension of pg.PlotDataItem with default qudi style plot options """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.opts['pen'] = _QudiPalette.c1
        self.opts['symbolPen'] = _QudiPalette.c1
        self.opts['symbolBrush'] = _QudiPalette.c1
        self.setData(*args, **kwargs)


class DataImageItem(_ImageItem):
    """ Extension of pg.ImageItem with percentile level scaling and image size adjustment """

    def __init__(self, image=None, **kwargs):
        # Change default color scale to qudi default
        if kwargs.get('lut', None) is None:
            kwargs['lut'] = _Colorscale().lut
        super().__init__(image, **kwargs)
        self._percentiles = None

    @property
    def percentiles(self) -> Union[None, Tuple[float, float]]:
        return self._percentiles

    def set_percentiles(self, percentiles: Union[None, Tuple[float, float]]) -> None:
        """ Set percentile range to clip image color level scaling.
        """
        if percentiles is not None:
            percentiles = (min(percentiles), max(percentiles))
        if percentiles != self._percentiles:
            self._percentiles = percentiles
            if self.image is not None:
                masked_image = np.ma.masked_invalid(self.image).compressed()
                if masked_image.size > 0:
                    self.setLevels(self._get_percentile_levels(masked_image))

    def set_image_extent(self,
                         extent: Tuple[Tuple[float, float], Tuple[float, float]],
                         adjust_for_px_size: Optional[bool] = True
                         ) -> None:
        """ Scales the image to a certain value range. By default, the resulting extent will be a
        bit larger, so that each pixel center corresponds to the respective xy coordinate.
        """
        if adjust_for_px_size is None:
            adjust_for_px_size = True
        if len(extent) != 2:
            raise ValueError('Image extent must be float sequence of length 2')
        if len(extent[0]) != 2 or len(extent[1]) != 2:
            raise TypeError('Image extent for each axis must be sequence of length 2.')

        if self.image is not None:
            x_min, x_max = min(extent[0]), max(extent[0])
            y_min, y_max = min(extent[1]), max(extent[1])
            if adjust_for_px_size:
                if self.image.shape[0] > 1 and self.image.shape[1] > 1:
                    half_px_x = (x_max - x_min) / (2 * (self.image.shape[0] - 1))
                    half_px_y = (y_max - y_min) / (2 * (self.image.shape[1] - 1))
                    x_min -= half_px_x
                    x_max += half_px_x
                    y_min -= half_px_y
                    y_max += half_px_y
            self.setRect(QtCore.QRectF(x_min, y_min, x_max - x_min, y_max - y_min))

    def set_image(self, image=None, **kwargs):
        """ vpg.ImageItem method override to apply optional filter when setting image data.
        """
        if image is None:
            self.clear()
        else:
            masked_image = np.ma.masked_invalid(image).compressed()
            if masked_image.size > 0:
                kwargs['levels'] = kwargs.get('levels', self._get_percentile_levels(masked_image))
                self.setImage(image=image, **kwargs)
            else:
                self.clear()

    def _get_percentile_levels(self, image) -> Tuple[float, float]:
        if self._percentiles is None:
            min_value = np.min(image)
            max_value = np.max(image)
        else:
            min_value = np.percentile(image, self._percentiles[0])
            max_value = np.percentile(image, self._percentiles[1])
        return min_value, max_value



# -*- coding: utf-8 -*-
"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ParameterEditor', 'ParameterEditorDialog']

import inspect
from typing import Any, Optional, Dict, Mapping, Callable
from PySide2 import QtCore, QtWidgets

from qudi.util.parameters import ParameterWidgetMapper


class ParameterEditor(QtWidgets.QWidget):
    """ Dynamically created editor widget for callable parameters.
    For best results use default values and simple type annotations in the callable to create the
    editor for.
    """
    INVALID = object()

    def __init__(self, *args, func: Callable, values: Optional[Mapping[str, Any]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        if values is None:
            values = dict()
        self.parameter_editors = dict()
        layout = QtWidgets.QGridLayout()
        parameters = inspect.signature(func).parameters
        for row, (name, param) in enumerate(parameters.items()):
            label = QtWidgets.QLabel(f'{name}:')
            label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            editor = ParameterWidgetMapper.widget_for_parameter(param)
            if editor is None:
                editor = QtWidgets.QLabel('Unknown argument type!')
                editor.setAlignment(QtCore.Qt.AlignCenter)
            else:
                editor = editor()
                # Attempt to set default value
                if not ((param.default is inspect.Parameter.empty) and (name not in values)):
                    try:
                        init_value = values[name]
                    except KeyError:
                        init_value = param.default

                    try:
                        editor.setValue(init_value)
                    except AttributeError:
                        try:
                            editor.setChecked(init_value)
                        except AttributeError:
                            try:
                                editor.setText(init_value)
                            except AttributeError:
                                pass

            layout.addWidget(label, row, 0)
            layout.addWidget(editor, row, 1)
            self.parameter_editors[name] = editor
        self.setLayout(layout)

    def get_parameter_values(self) -> Dict[str, Any]:
        """ Returns the current parameter values entered into the editor """
        values = dict()
        for name, editor in self.parameter_editors.items():
            if isinstance(editor, QtWidgets.QLabel):
                values[name] = self.INVALID
            else:
                try:
                    values[name] = editor.value()
                except AttributeError:
                    try:
                        values[name] = editor.isChecked()
                    except AttributeError:
                        try:
                            values[name] = editor.text()
                        except AttributeError:
                            pass
        return values


class ParameterEditorDialog(QtWidgets.QDialog):
    """ QDialog containing a ParameterEditor widget and OK, Cancel and Apply buttons.
    """
    def __init__(self, *args, func: Callable, values: Optional[Mapping[str, Any]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.parameter_editor = ParameterEditor(func=func, values=values)
        self.scroll_area = QtWidgets.QScrollArea()
        self.scroll_area.setWidget(self.parameter_editor)
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok |
            QtWidgets.QDialogButtonBox.Cancel |
            QtWidgets.QDialogButtonBox.Apply,
            orientation=QtCore.Qt.Horizontal
        )
        self.ok_button = self.button_box.button(QtWidgets.QDialogButtonBox.Ok)
        self.cancel_button = self.button_box.button(QtWidgets.QDialogButtonBox.Cancel)
        self.apply_button = self.button_box.button(QtWidgets.QDialogButtonBox.Apply)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.scroll_area)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

        # Connect signals
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)



# -*- coding: utf-8 -*-

"""
This file contains unit tests for all qudi fit routines for exponential decay models.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import unittest
import numpy as np

from qudi.util.fit_models.sine import Sine, DoubleSine


class TestSineMethods(unittest.TestCase):
    _fit_param_tolerance = 0.05  # 5% tolerance for each fit parameter

    @staticmethod
    def sine(x, offset, amplitude, frequency, phase):
        return offset + amplitude * np.sin(2 * np.pi * frequency * x + phase)

    def setUp(self):
        self.offset = (np.random.rand() - 0.5) * 2e6
        self.amplitudes = np.random.rand(3) * 100
        window = max(1e-9, np.random.rand() * 1e9)
        left = (np.random.rand() - 0.5) * 2e9
        points = np.random.randint(10, 1001)
        self.x_values = np.linspace(left, left + window, points)
        max_freq = 0.25 / (self.x_values[1] - self.x_values[0])
        min_freq = 0.5 / window
        self.frequencies = min_freq + np.random.rand(3) * (max_freq - min_freq)
        self.phases = (np.random.rand(3) - 0.5) * 2 * np.pi
        self.noise_amp = max(self.amplitudes.min() / 10, np.random.rand() * self.amplitudes.min())
        self.noise = (np.random.rand(points) - 0.5) * self.noise_amp

    def test_sine(self):
        # Test for sine fit
        y_values = self.noise + self.sine(self.x_values,
                                          self.offset,
                                          self.amplitudes[0],
                                          self.frequencies[0],
                                          self.phases[0])

        fit_model = Sine()
        fit_result = fit_model.fit(data=y_values,
                                   x=self.x_values,
                                   **fit_model.guess(y_values, self.x_values))

        params_ideal = {'offset': self.offset,
                        'amplitude': self.amplitudes[0],
                        'frequency': self.frequencies[0],
                        'phase': self.phases[0]}
        for name, ideal_val in params_ideal.items():
            diff = abs(fit_result.best_values[name] - ideal_val)
            tolerance = abs(ideal_val * self._fit_param_tolerance)
            msg = 'Sine fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)

    def test_double_sine(self):
        # Test for sine fit
        y_values = self.noise + self.sine(self.x_values,
                                          self.offset,
                                          self.amplitudes[0],
                                          self.frequencies[0],
                                          self.phases[0])
        y_values += self.sine(self.x_values,
                              self.offset,
                              self.amplitudes[1],
                              self.frequencies[1],
                              self.phases[1])

        fit_model = DoubleSine()
        fit_result = fit_model.fit(data=y_values,
                                   x=self.x_values,
                                   **fit_model.guess(y_values, self.x_values))

        params_ideal = {'offset': self.offset,
                        'amplitude_1': self.amplitudes[0],
                        'amplitude_2': self.amplitudes[1],
                        'frequency_1': self.frequencies[0],
                        'frequency_2': self.frequencies[0],
                        'phase_1': self.phases[0],
                        'phase_2': self.phases[1]}
        for name, ideal_val in params_ideal.items():
            diff = abs(fit_result.best_values[name] - ideal_val)
            tolerance = abs(ideal_val * self._fit_param_tolerance)
            msg = 'Double sine fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['csv_2_list', 'in_range', 'is_complex', 'is_complex_type', 'is_float', 'is_float_type',
           'is_integer', 'is_integer_type', 'is_number', 'is_number_type', 'is_string',
           'is_string_type', 'iter_modules_recursive', 'natural_sort', 'str_to_number']

import re
import os
import pkgutil
import numpy as np
from typing import Union, Optional, Iterable, List, Any, Type, Tuple, Callable

_RealNumber = Union[int, float]


def iter_modules_recursive(paths: Union[str, Iterable[str]],
                           prefix: Optional[str] = '') -> List[pkgutil.ModuleInfo]:
    """Has the same signature as pkgutil.iter_modules() but extends the functionality by walking
    through the entire directory tree and concatenating the return values of pkgutil.iter_modules()
    for each directory.

    Additional modifications include:
    - Directories starting with "_" or "." are ignored (including their sub-directories).
    - Python modules starting with a double-underscore ("__") are excluded from the result.

    Parameters
    ----------
    paths : iterable
        Iterable of root directories to start the search for modules.
    prefix : str, optional
        Prefix to prepend to all module names.

    Returns
    -------
    iterable
        Concatenated return values of pkgutil.iter_modules() for all directories in the tree.
    """
    if isinstance(paths, str):
        paths = [paths]
    module_infos = list()
    for search_top in paths:
        for root, dirs, files in os.walk(search_top):
            rel_path = os.path.relpath(root, search_top)
            if rel_path and rel_path != '.' and rel_path[0] in '._':
                # Prevent os.walk to descent further down this tree branch
                dirs.clear()
                # Ignore this directory
                continue
            # Resolve current module prefix
            if not rel_path or rel_path == '.':
                curr_prefix = prefix
            else:
                curr_prefix = prefix + '.'.join(rel_path.split(os.sep)) + '.'
            # find modules and packages in current dir
            tmp = pkgutil.iter_modules([root], prefix=curr_prefix)
            module_infos.extend(
                [mod_inf for mod_inf in tmp if not mod_inf.name.rsplit('.', 1)[-1].startswith('__')]
            )
    return module_infos


def natural_sort(iterable: Iterable[Any]) -> List[Any]:
    """
    Sort an iterable of strings in an intuitive, natural way (human/natural sort).
    This is useful for sorting alphanumeric strings that contain integers.

    Parameters
    ----------
    iterable : list of str
        Iterable with string items to sort.

    Returns
    -------
    list
        Sorted list of strings.
    """
    def conv(s):
        return int(s) if s.isdigit() else s
    try:
        return sorted(iterable, key=lambda key: [conv(i) for i in re.split(r'(\d+)', key)])
    except:
        return sorted(iterable)


def is_number(test_value: Any) -> bool:
    """Check whether passed value is a number."""
    return is_integer(test_value) or is_float(test_value) or is_complex(test_value)


def is_number_type(test_obj: Type) -> bool:
    """Check whether passed object is a number type."""
    return is_integer_type(test_obj) or is_float_type(test_obj) or is_complex_type(test_obj)


def is_integer(test_value: Any) -> bool:
    """Check all available integer representations."""
    return isinstance(test_value, (int, np.integer))


def is_integer_type(test_obj: Type) -> bool:
    """Check if passed object is an integer type."""
    return issubclass(test_obj, (int, np.integer))


def is_float(test_value: Any) -> bool:
    """Check all available float representations."""
    return isinstance(test_value, (float, np.floating))


def is_float_type(test_obj: Type) -> bool:
    """Check if passed object is a float type."""
    return issubclass(test_obj, (float, np.floating))


def is_complex(test_value: Any) -> bool:
    """Check all available complex representations."""
    return isinstance(test_value, (complex, np.complexfloating))


def is_complex_type(test_obj: Type) -> bool:
    """Check if passed object is a complex type."""
    return issubclass(test_obj, (complex, np.complexfloating))


def is_string(test_value: Any) -> bool:
    """Check all available string representations."""
    return isinstance(test_value, (str, np.str_, np.string_))


def is_string_type(test_obj: Type) -> bool:
    """Check if passed object is a string type."""
    return issubclass(test_obj, (str, np.str_, np.string_))


def in_range(value: _RealNumber, lower_limit: _RealNumber,
             upper_limit: _RealNumber) -> Tuple[bool, _RealNumber]:
    """Check if a value is in a given range an return closest possible value in range.
    Also check the range.
    Return value is clipped to range.
    """
    if upper_limit < lower_limit:
        lower_limit, upper_limit = upper_limit, lower_limit

    if value > upper_limit:
        return False, upper_limit
    if value < lower_limit:
        return False, lower_limit
    return True, value


def csv_2_list(csv_string: str, str_2_val: Optional[Callable[[str], Any]] = None) -> List[Any]:
    """
    Parse a list literal (with or without square brackets) given as a string containing
    comma-separated int or float values to a Python list.
    
    Blanks before and after commas are handled.

    Parameters
    ----------
    csv_string : str
        Scalar number literals as strings separated by a single comma and any number
        of blanks. Brackets are ignored.
        Example: '[1e-6,2.5e6, 42]' or '1e-6, 2e-6,   42'.
    str_2_val : function, optional
        Function to use for casting substrings into single values.

    Returns
    -------
    list
        List of float values. If `str_2_val` is provided, type is invoked by this function.
    """
    if not isinstance(csv_string, str):
        raise TypeError('string_2_list accepts only str type input.')

    if csv_string == "":
        return []

    csv_string = csv_string.replace('[', '').replace(']', '')  # Remove square brackets
    csv_string = csv_string.replace('(', '').replace(')', '')  # Remove round brackets
    csv_string = csv_string.replace('{', '').replace('}', '')  # Remove curly brackets
    csv_string = csv_string.strip().strip(',')  # Remove trailing/leading blanks and commas

    # Cast each str value to float if no explicit cast function is given by parameter str_2_val.
    if str_2_val is None:
        csv_list = [str_to_number(val_str) for val_str in csv_string.split(',')]
    else:
        csv_list = [str_2_val(val_str.strip()) for val_str in csv_string.split(',')]
    return csv_list


def str_to_number(str_value: str,
                  return_failed: Optional[bool] = False) -> Union[int, float, complex, str]:
    """Parse a string into either int, float or complex (in that order)."""
    try:
        return int(str_value)
    except ValueError:
        try:
            return float(str_value)
        except ValueError:
            try:
                return complex(str_value)
            except ValueError:
                if return_failed:
                    return str_value
                else:
                    raise ValueError(
                        f'Could not convert string to int, float or complex: \'{str_value}\''
                    )



# -*- coding: utf-8 -*-

"""
This file contains models of Sine fitting routines for qudi based on the lmfit package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('Sine', 'DoubleSine', 'ExponentialDecaySine', 'estimate_frequency_ft')

import numpy as np
from qudi.util.math import compute_ft
from qudi.util.fit_models.model import FitModelBase, estimator
from qudi.util.fit_models.helpers import sort_check_data


def estimate_frequency_ft(data, x):
    # calculate PSD with zeropadding to obtain nicer interpolation between the appearing peaks.
    dft_x, dft_y = compute_ft(x, data, zeropad_num=1, psd=True)
    # Maximum PSD value corresponds to most likely frequency
    return abs(dft_x[dft_y.argmax()]), (dft_x, dft_y)


class Sine(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=1., min=0., max=np.inf)
        self.set_param_hint('frequency', value=0., min=0., max=np.inf)
        self.set_param_hint('phase', value=0., min=-np.pi, max=np.pi)

    @staticmethod
    def _model_function(x, offset, amplitude, frequency, phase):
        return offset + amplitude * np.sin(2 * np.pi * frequency * x + phase)

    @estimator('default')
    def estimate(self, data, x):
        data, x = sort_check_data(data, x)
        x_span = abs(max(x) - min(x))
        offset = np.mean(data)

        estimate = self.estimate_no_offset(data - offset, x)
        if 1/(2 * estimate['frequency'].value) > x_span:
            estimate['offset'].set(value=offset, min=-np.inf, max=np.inf, vary=True)
        else:
            estimate['offset'].set(value=offset, min=min(data), max=max(data), vary=True)
        return estimate

    @estimator('No Offset')
    def estimate_no_offset(self, data, x):
        data, x = sort_check_data(data, x)
        x_step = min(abs(np.ediff1d(x)))
        data_span = abs(max(data) - min(data))
        amplitude = data_span / 2

        frequency, _ = estimate_frequency_ft(data, x)

        # Find an estimate for the phase
        # Procedure: Create sin waves with different phases and perform a summation.
        #            The sum shows how well the sine was fitting to the actual data.
        #            The best fitting sine should be a maximum of the summed time
        #            trace.
        iter_steps = max(1, int(round(1 / (frequency * x_step))))
        test_phases = 2 * np.pi * np.arange(iter_steps) / iter_steps
        sum_res = np.zeros(iter_steps)
        for ii, phase in enumerate(test_phases):
            sum_res[ii] = np.abs(data - amplitude * np.sin(2 * np.pi * frequency * x + phase)).sum()
        phase = test_phases[sum_res.argmax()] - np.pi  # Maximum sum value corresponds to worst fit

        estimate = self.make_params()
        estimate['frequency'].set(value=frequency, min=0, max=1 / (2 * x_step), vary=True)
        estimate['amplitude'].set(value=amplitude, min=0, max=2 * data_span, vary=True)
        estimate['phase'].set(value=phase, min=-np.pi, max=np.pi, vary=True)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Zero Phase')
    def estimate_zero_phase(self, data, x):
        data, x = sort_check_data(data, x)
        x_step = min(abs(np.ediff1d(x)))
        x_span = abs(max(x) - min(x))
        data_span = abs(max(data) - min(data))

        amplitude = data_span / 2
        offset = np.mean(data)
        frequency, _ = estimate_frequency_ft(data - offset, x)

        estimate = self.make_params()
        estimate['frequency'].set(value=frequency, min=0, max=1 / (2 * x_step), vary=True)
        estimate['amplitude'].set(value=amplitude, min=0, max=2 * data_span, vary=True)
        if 1/(2 * frequency) > x_span:
            estimate['offset'].set(value=offset, min=-np.inf, max=np.inf, vary=True)
        else:
            estimate['offset'].set(value=offset, min=min(data), max=max(data), vary=True)
        estimate['phase'].set(value=0, min=-np.pi, max=np.pi, vary=False)
        return estimate


class DoubleSine(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=1., min=0., max=np.inf)
        self.set_param_hint('amplitude_2', value=1., min=0., max=np.inf)
        self.set_param_hint('frequency_1', value=0., min=0., max=np.inf)
        self.set_param_hint('frequency_2', value=0., min=0., max=np.inf)
        self.set_param_hint('phase_1', value=0., min=-np.pi, max=np.pi)
        self.set_param_hint('phase_2', value=0., min=-np.pi, max=np.pi)

    @staticmethod
    def _model_function(x, offset, amplitude_1, amplitude_2, frequency_1, frequency_2, phase_1,
                        phase_2):
        result = amplitude_1 * np.sin(2 * np.pi * frequency_1 * x + phase_1)
        result += amplitude_2 * np.sin(2 * np.pi * frequency_2 * x + phase_2)
        return result + offset

    @estimator('default')
    def estimate(self, data, x):
        x_span = abs(max(x) - min(x))
        offset = np.mean(data)

        estimate = self.estimate_no_offset(data - offset, x)
        if 1 / (2 * min(estimate['frequency_1'].value, estimate['frequency_2'].value)) > x_span:
            estimate['offset'].set(value=offset, min=-np.inf, max=np.inf, vary=True)
        else:
            estimate['offset'].set(value=offset, min=min(data), max=max(data), vary=True)
        return estimate

    @estimator('No Offset')
    def estimate_no_offset(self, data, x):
        data, x = sort_check_data(data, x)
        # Fit a single sine to the data
        single_sine_model = Sine()
        single_sine_result = single_sine_model.fit(data,
                                                   single_sine_model.estimate_no_offset(data, x),
                                                   x=x)
        # Subtract the fitted sine and estimate another single sine from the remaining data
        data_sub = data - single_sine_result.best_fit
        single_sine_estimate = single_sine_model.estimate_no_offset(data_sub, x)
        # Merge single sine fit result and single sine estimate

        single_fit_params = single_sine_result.params
        estimate = self.make_params()
        estimate['amplitude_1'].set(value=single_fit_params['amplitude'].value,
                                    min=single_fit_params['amplitude'].min,
                                    max=single_fit_params['amplitude'].max,
                                    vary=True)
        estimate['amplitude_2'].set(value=single_sine_estimate['amplitude'].value,
                                    min=single_sine_estimate['amplitude'].min,
                                    max=single_sine_estimate['amplitude'].max,
                                    vary=True)
        estimate['frequency_1'].set(value=single_fit_params['frequency'].value,
                                    min=single_fit_params['frequency'].min,
                                    max=single_fit_params['frequency'].max,
                                    vary=True)
        estimate['frequency_2'].set(value=single_sine_estimate['frequency'].value,
                                    min=single_sine_estimate['frequency'].min,
                                    max=single_sine_estimate['frequency'].max,
                                    vary=True)
        estimate['phase_1'].set(value=single_fit_params['phase'].value,
                                min=single_fit_params['phase'].min,
                                max=single_fit_params['phase'].max,
                                vary=True)
        estimate['phase_2'].set(value=single_sine_estimate['phase'].value,
                                min=single_sine_estimate['phase'].min,
                                max=single_sine_estimate['phase'].max,
                                vary=True)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate


class ExponentialDecaySine(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=1., min=0., max=np.inf)
        self.set_param_hint('frequency', value=0., min=0., max=np.inf)
        self.set_param_hint('phase', value=0., min=-np.pi, max=np.pi)
        self.set_param_hint('decay', value=1., min=0., max=np.inf)
        self.set_param_hint('stretch', value=1., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, amplitude, frequency, phase, decay, stretch):
        return offset + amplitude * np.exp(-(x / decay) ** stretch) * np.sin(
            2 * np.pi * frequency * x + phase)

    @estimator('Decay')
    def estimate_decay(self, data, x):
        x_span = abs(max(x) - min(x))
        offset = np.mean(data)

        estimate = self.estimate_decay_no_offset(data - offset, x)
        if 1 / (2 * estimate['frequency'].value) > x_span:
            estimate['offset'].set(value=offset, min=-np.inf, max=np.inf, vary=True)
        else:
            estimate['offset'].set(value=offset, min=min(data), max=max(data), vary=True)
        return estimate

    @estimator('Stretched Decay')
    def estimate_stretched_decay(self, data, x):
        # ToDo: Stretch estimation
        estimate = self.estimate_decay(data, x)
        estimate['stretch'].set(value=2, min=0, max=np.inf, vary=True)
        return estimate

    @estimator('Decay (no offset)')
    def estimate_decay_no_offset(self, data, x):
        data, x = sort_check_data(data, x)
        x_step = min(abs(np.ediff1d(x)))
        data_span = abs(max(data) - min(data))
        amplitude = data_span / 2

        frequency, (dft_x, dft_y) = estimate_frequency_ft(data, x)

        # remove noise for peak width and decay constant estimation
        dft_y[np.argwhere(dft_y <= np.std(dft_y))] = 0
        # calculating the width of the FT peak for the estimation of decay constant
        decay = 1 / (2 * np.trapz(dft_y, dft_x) / max(dft_y))
        # s = 0
        # for i in range(0, len(dft_x)):
        #     s += dft_y[i] * abs(dft_x[1] - dft_x[0]) / max(dft_y)
        # lifetime_val = 0.5 / s

        # Find an estimate for the phase
        # Procedure: Create sin waves with different phases and perform a summation.
        #            The sum shows how well the sine was fitting to the actual data.
        #            The best fitting sine should be a maximum of the summed time
        #            trace.
        iter_steps = max(1, int(round(1 / (frequency * x_step))))
        test_phases = 2 * np.pi * np.arange(iter_steps) / iter_steps
        sum_res = np.zeros(iter_steps)
        for ii, phase in enumerate(test_phases):
            sum_res[ii] = np.abs(data - amplitude * np.sin(2 * np.pi * frequency * x + phase)).sum()
        phase = test_phases[sum_res.argmax()] - np.pi  # Maximum sum value corresponds to worst fit

        estimate = self.make_params()
        estimate['frequency'].set(value=frequency, min=0, max=1 / (2 * x_step), vary=True)
        estimate['amplitude'].set(value=amplitude, min=0, max=2 * data_span, vary=True)
        estimate['phase'].set(value=phase, min=-np.pi, max=np.pi, vary=True)
        estimate['decay'].set(value=decay,
                              min=2 * x_step,
                              max=1/(abs(dft_x[1]-dft_x[0])*0.5),
                              vary=True)
        estimate['stretch'].set(value=1, min=0, max=np.inf, vary=False)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Stretched Decay (no offset)')
    def estimate_stretched_decay_no_offset(self, data, x):
        # ToDo: Stretch estimation
        estimate = self.estimate_decay_no_offset(data, x)
        estimate['stretch'].set(value=2, min=0, max=np.inf, vary=True)
        return estimate


class ExponentialDecayDoubleSine(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=1., min=0., max=np.inf)
        self.set_param_hint('amplitude_2', value=1., min=0., max=np.inf)
        self.set_param_hint('frequency_1', value=0., min=0., max=np.inf)
        self.set_param_hint('frequency_2', value=0., min=0., max=np.inf)
        self.set_param_hint('phase_1', value=0., min=-np.pi, max=np.pi)
        self.set_param_hint('phase_2', value=0., min=-np.pi, max=np.pi)
        self.set_param_hint('decay', value=1., min=0., max=np.inf)
        self.set_param_hint('stretch', value=1., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, amplitude_1, amplitude_2, frequency_1, frequency_2, phase_1,
                        phase_2, decay, stretch):
        result = amplitude_1 * np.sin(2 * np.pi * frequency_1 * x + phase_1)
        result += amplitude_2 * np.sin(2 * np.pi * frequency_2 * x + phase_2)
        return np.exp(-(x / decay) ** stretch) * result + offset

    @estimator('Decay')
    def estimate_decay(self, data, x):
        x_span = abs(max(x) - min(x))
        offset = np.mean(data)

        estimate = self.estimate_decay_no_offset(data - offset, x)
        if 1 / (2 * min(estimate['frequency_1'].value, estimate['frequency_2'].value)) > x_span:
            estimate['offset'].set(value=offset, min=-np.inf, max=np.inf, vary=True)
        else:
            estimate['offset'].set(value=offset, min=min(data), max=max(data), vary=True)
        return estimate

    @estimator('Stretched Decay')
    def estimate_stretched_decay(self, data, x):
        # ToDo: Stretch estimation
        estimate = self.estimate_decay(data, x)
        estimate['stretch'].set(value=2, min=0, max=np.inf, vary=True)
        return estimate

    @estimator('Decay (no offset)')
    def estimate_decay_no_offset(self, data, x):
        data, x = sort_check_data(data, x)

        # Try to estimate fit from fitting single sine cases
        model = ExponentialDecaySine()
        first_sine_fit = model.fit(data, model.estimate_decay_no_offset(data, x), x=x)
        data_sub = data - first_sine_fit.best_fit
        model = Sine()
        second_sine_fit = model.fit(data_sub, model.estimate_no_offset(data_sub, x), x=x)

        # Merge fit results into estimated parameters
        first_params = first_sine_fit.params
        second_params = second_sine_fit.params
        estimate = self.make_params()
        estimate['frequency_1'].set(value=first_params['frequency'].value,
                                    min=first_params['frequency'].min,
                                    max=first_params['frequency'].max,
                                    vary=True)
        estimate['frequency_2'].set(value=second_params['frequency'].value,
                                    min=second_params['frequency'].min,
                                    max=second_params['frequency'].max,
                                    vary=True)
        estimate['amplitude_1'].set(value=first_params['amplitude'].value,
                                    min=first_params['amplitude'].min,
                                    max=first_params['amplitude'].max,
                                    vary=True)
        estimate['amplitude_2'].set(value=second_params['amplitude'].value,
                                    min=second_params['amplitude'].min,
                                    max=second_params['amplitude'].max,
                                    vary=True)
        estimate['phase_1'].set(value=first_params['phase'].value,
                                min=first_params['phase'].min,
                                max=first_params['phase'].max,
                                vary=True)
        estimate['phase_2'].set(value=second_params['phase'].value,
                                min=second_params['phase'].min,
                                max=second_params['phase'].max,
                                vary=True)
        estimate['decay'].set(value=first_params['decay'].value,
                              min=first_params['decay'].min,
                              max=first_params['decay'].max,
                              vary=True)
        estimate['stretch'].set(value=1, min=0, max=np.inf, vary=False)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Stretched Decay (no offset)')
    def estimate_stretched_decay_no_offset(self, data, x):
        # ToDo: Stretch estimation
        estimate = self.estimate_decay_no_offset(data, x)
        estimate['stretch'].set(value=2, min=0, max=np.inf, vary=True)
        return estimate


# class DoubleExponentialDecayDoubleSine(FitModelBase):
#     """
#     """
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
#         self.set_param_hint('amplitude_1', value=1., min=0., max=np.inf)
#         self.set_param_hint('amplitude_2', value=1., min=0., max=np.inf)
#         self.set_param_hint('frequency_1', value=0., min=0., max=np.inf)
#         self.set_param_hint('frequency_2', value=0., min=0., max=np.inf)
#         self.set_param_hint('phase_1', value=0., min=-np.pi, max=np.pi)
#         self.set_param_hint('phase_2', value=0., min=-np.pi, max=np.pi)
#         self.set_param_hint('decay_1', value=1., min=0., max=np.inf)
#         self.set_param_hint('decay_2', value=1., min=0., max=np.inf)
#
#     @staticmethod
#     def _model_function(x, offset, amplitude_1, amplitude_2, frequency_1, frequency_2, phase_1,
#                         phase_2, decay_1, decay_2):
#         result = np.exp(-x / decay_1) * amplitude_1 * np.sin(2 * np.pi * frequency_1 * x + phase_1)
#         result += np.exp(-x / decay_2) * amplitude_2 * np.sin(2 * np.pi * frequency_2 * x + phase_2)
#         return result + offset



# -*- coding: utf-8 -*-
"""
This file contains the qudi task runner GUI.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtCore

from qudi.core.connector import Connector
from qudi.core.module import GuiBase

from .main_window import TaskMainWindow


class TaskRunnerGui(GuiBase):
    """
    TODO: Document
    """

    # declare connectors
    _task_runner = Connector(name='task_runner', interface='TaskRunnerLogic')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mw = None

    def on_activate(self):
        """Create all UI objects and show the window.
        """
        # Initialize main window and connect task widgets
        taskrunner = self._task_runner()
        self._mw = TaskMainWindow(tasks=taskrunner.configured_task_types)
        self._mw.sigStartTask.connect(taskrunner.run_task, QtCore.Qt.QueuedConnection)
        self._mw.sigInterruptTask.connect(taskrunner.interrupt_task, QtCore.Qt.QueuedConnection)
        self._mw.sigClosed.connect(self._deactivate_self)
        taskrunner.sigTaskStarted.connect(self._mw.task_started, QtCore.Qt.QueuedConnection)
        taskrunner.sigTaskStateChanged.connect(self._mw.task_state_changed,
                                               QtCore.Qt.QueuedConnection)
        taskrunner.sigTaskFinished.connect(self._mw.task_finished, QtCore.Qt.QueuedConnection)

        # Set current task states
        for task_name, task_state in taskrunner.task_states.items():
            if task_state != 'stopped':
                self._mw.task_started(task_name)
            else:
                self._mw.task_finished(task_name, None, False)
            self._mw.task_state_changed(task_name, task_state)
        # ToDo: Also set task results here

        self._restore_window_geometry(self._mw)
        self.show()

    def show(self):
        """Make sure that the window is visible and at the top.
        """
        self._mw.show()

    @QtCore.Slot()
    def _deactivate_self(self):
        self._qudi_main.module_manager.deactivate_module(self._meta['name'])

    def on_deactivate(self):
        """Hide window and stop ipython console.
        """
        self._save_window_geometry(self._mw)
        self._mw.close()
        self._mw.sigStartTask.disconnect()
        self._mw.sigInterruptTask.disconnect()
        self._mw.sigClosed.disconnect()
        taskrunner = self._task_runner()
        taskrunner.sigTaskStarted.disconnect(self._mw.task_started, QtCore.Qt.QueuedConnection)
        taskrunner.sigTaskStateChanged.disconnect(self._mw.task_state_changed,
                                               QtCore.Qt.QueuedConnection)
        taskrunner.sigTaskFinished.disconnect(self._mw.task_finished, QtCore.Qt.QueuedConnection)
        self._mw = None



# -*- coding: utf-8 -*-
"""
This file contains the qudi tools for remote module sharing via rpyc server.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('RemoteModulesService', 'QudiNamespaceService')

import logging

import rpyc
import weakref
from functools import wraps
from inspect import signature, isfunction, ismethod

from qudi.util.mutex import Mutex
from qudi.util.models import DictTableModel
from qudi.util.network import netobtain
from qudi.core.logger import get_logger

logger = get_logger(__name__)


class _SharedModulesModel(DictTableModel):
    """Derived dict model for GUI display elements.
    """
    def __init__(self):
        super().__init__(headers='Shared Module')

    def data(self, index, role):
        """Get data from model for a given cell. Data can have a role that affects display.

        Parameters
        ----------
        index : QModelIndex
            Cell for which data is requested.
        role : ItemDataRole
            Role for which data is requested.

        Returns
        -------
        QVariant
            Data for the given cell and role.
        """
        data = super().data(index, role)
        if data is None:
            return None
        # second column returns weakref.ref object
        if index.column() == 1:
            data = data()
        return data


class RemoteModulesService(rpyc.Service):
    """An RPyC service that has a module list.
    """
    ALIASES = ['RemoteModules']

    def __init__(self, *args, force_remote_calls_by_value=False, **kwargs):
        super().__init__(*args, **kwargs)
        self._thread_lock = Mutex()
        self.shared_modules = _SharedModulesModel()
        self._force_remote_calls_by_value = force_remote_calls_by_value

    def share_module(self, module):
        with self._thread_lock:
            if module.name in self.shared_modules:
                logger.warning(f'Module "{module.name}" already shared')
                return
            self.shared_modules[module.name] = weakref.ref(module)
            weakref.finalize(module, self.remove_shared_module, module.name)

    def remove_shared_module(self, module):
        with self._thread_lock:
            name = module if isinstance(module, str) else module.name
            self.shared_modules.pop(name, None)

    def on_connect(self, conn):
        """Code that runs when a connection is created.
        """
        host, port = conn._config['endpoints'][1]
        logger.info(f'Client connected to remote modules service from [{host}]:{port:d}')

    def on_disconnect(self, conn):
        """Code that runs when the connection is closing.
        """
        host, port = conn._config['endpoints'][1]
        logger.info(f'Client [{host}]:{port:d} disconnected from remote modules service')

    def exposed_get_module_instance(self, name, activate=False):
        """Return reference to a module in the shared module list.

        Parameters
        ----------
        name : str
            Unique module name.

        Returns
        -------
        object
            Reference to the module.
        """
        with self._thread_lock:
            try:
                module = self.shared_modules.get(name, None)()
            except TypeError:
                logger.error(f'Client requested a module ("{name}") that is not shared.')
                return None
            if activate:
                if not module.activate():
                    logger.error(f'Unable to share requested module "{name}" with client. Module '
                                 f'can not be activated.')
                    return None
            if self._force_remote_calls_by_value:
                return ModuleRpycProxy(module.instance)
            return module.instance

    def exposed_get_available_module_names(self):
        """Returns the currently shared module names independent of the current module state.

        Returns
        -------
        tuple
            Names of the currently shared modules.
        """
        with self._thread_lock:
            return tuple(self.shared_modules)

    def exposed_get_loaded_module_names(self):
        """Returns the currently shared module names for all modules that have been loaded
        (instantiated).

        Returns
        -------
        tuple
            Names of the currently shared and loaded modules.
        """
        with self._thread_lock:
            all_modules = {name: ref() for name, ref in self.shared_modules.items()}
            return tuple(name for name, mod in all_modules.items() if
                         mod is not None and mod.instance is not None)

    def exposed_get_active_module_names(self):
        """Returns the currently shared module names for all modules that are active.

        Returns
        -------
        tuple
            Names of the currently shared active modules.
        """
        with self._thread_lock:
            all_modules = {name: ref() for name, ref in self.shared_modules.items()}
            return tuple(
                name for name, mod in all_modules.items() if mod is not None and mod.is_active
            )


class QudiNamespaceService(rpyc.Service):
    """An RPyC service providing a namespace dict containing references to all active qudi module
    instances as well as a reference to the qudi application itself.
    """
    ALIASES = ['QudiNamespace']

    def __init__(self, *args, qudi, force_remote_calls_by_value=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.__qudi_ref = weakref.ref(qudi)
        self._notifier_callbacks = dict()
        self._force_remote_calls_by_value = force_remote_calls_by_value

    @property
    def _qudi(self):
        qudi = self.__qudi_ref()
        if qudi is None:
            raise RuntimeError('Dead qudi application reference encountered')
        return qudi

    @property
    def _module_manager(self):
        manager = self._qudi.module_manager
        if manager is None:
            raise RuntimeError('No module manager initialized in qudi application')
        return manager

    def on_connect(self, conn):
        """Code that runs when a connection is created.
        """
        try:
            self._notifier_callbacks[conn] = rpyc.async_(conn.root.modules_changed)
        except AttributeError:
            pass
        host, port = conn._config['endpoints'][1]
        logger.info(f'Client connected to local module service from [{host}]:{port:d}')

    def on_disconnect(self, conn):
        """Code that runs when the connection is closing.
        """
        self._notifier_callbacks.pop(conn, None)
        host, port = conn._config['endpoints'][1]
        logger.info(f'Client [{host}]:{port:d} disconnected from local module service')

    def notify_module_change(self):
        logger.debug('Local module server has detected a module state change and sends async '
                     'notifier signals to all clients')
        for callback in self._notifier_callbacks.values():
            callback()

    def exposed_get_namespace_dict(self):
        """Returns the instances of the currently active modules as well as a reference to the
        qudi application itself.

        Returns
        -------
        dict
            Names (keys) and object references (values).
        """
        if self._force_remote_calls_by_value:
            mods = {name: ModuleRpycProxy(mod.instance) for name, mod in
                    self._module_manager.items() if mod.is_active}
        else:
            mods = {name: mod.instance for name, mod in self._module_manager.items() if
                    mod.is_active}
        mods['qudi'] = self._qudi
        return mods

    def exposed_get_logger(self, name: str) -> logging.Logger:
        """Returns a logger object for remote processes to log into the qudi logging facility."""
        return get_logger(name)


class ModuleRpycProxy:
    """Instances of this class serve as proxies for qudi modules accessed via RPyC.
    It currently wraps all API methods (none- and single-underscore methods) to only receive
    parameters "by value", i.e. using qudi.util.network.netobtain. This will only work if all
    method arguments are "pickle-able".
    In addition all values passed to __setattr__ are also received "by value".

    Proxy class concept heavily inspired by this python recipe under PSF License:
    https://code.activestate.com/recipes/496741-object-proxying/
    """

    __slots__ = ['_obj_ref', '__weakref__']

    def __init__(self, obj):
        object.__setattr__(self, '_obj_ref', weakref.ref(obj))

    # proxying (special cases)
    def __getattribute__(self, name):
        obj = object.__getattribute__(self, '_obj_ref')()
        attr = getattr(obj, name)
        if not name.startswith('__') and ismethod(attr) or isfunction(attr):
            sig = signature(attr)
            if len(sig.parameters) > 0:

                @wraps(attr)
                def wrapped(*args, **kwargs):
                    sig.bind(*args, **kwargs)
                    args = [netobtain(arg) for arg in args]
                    kwargs = {name: netobtain(arg) for name, arg in kwargs.items()}
                    return attr(*args, **kwargs)

                wrapped.__signature__ = sig
                return wrapped
        return attr

    def __delattr__(self, name):
        obj = object.__getattribute__(self, '_obj_ref')()
        return delattr(obj, name)

    def __setattr__(self, name, value):
        obj = object.__getattribute__(self, '_obj_ref')()
        return setattr(obj, name, netobtain(value))

    # factories
    _special_names = (
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', '__contains__',
        '__delitem__', '__delslice__', '__div__', '__divmod__', '__eq__', '__float__',
        '__floordiv__', '__ge__', '__getitem__', '__getslice__', '__gt__', '__hash__', '__hex__',
        '__iadd__', '__iand__', '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__',
        '__imod__', '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', '__long__',
        '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__', '__neg__', '__oct__', '__or__',
        '__pos__', '__pow__', '__radd__', '__rand__', '__rdiv__', '__rdivmod__', '__reduce__',
        '__reduce_ex__', '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__',
        '__rxor__', '__setitem__', '__setslice__', '__sub__', '__truediv__', '__xor__', 'next',
        '__str__', '__nonzero__'
    )

    @classmethod
    def _create_class_proxy(cls, theclass):
        """Creates a proxy for the given class.
        """

        def make_method(method_name):

            def method(self, *args, **kw):
                obj = object.__getattribute__(self, '_obj_ref')()
                args = [netobtain(arg) for arg in args]
                kw = {key: netobtain(val) for key, val in kw.items()}
                return getattr(obj, method_name)(*args, **kw)

            return method

        # Add all special names to this wrapper class if they are present in the original class
        namespace = dict()
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)

        return type(f'{cls.__name__}({theclass.__name__})', (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        """Creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are passed to this
        class' __init__, so deriving classes can define an __init__ method of their own.

        note: _class_proxy_cache is unique per class (each deriving class must hold its own cache).
        """
        theclass = cls._create_class_proxy(obj.__class__)
        return object.__new__(theclass)



# -*- coding: utf-8 -*-
"""

"""

__all__ = ['AvailableModulesTreeWidget', 'SelectedModulesTreeWidget', 'ConfigModulesTreeWidget']

from re import compile
from PySide2 import QtCore, QtWidgets, QtGui
from typing import Optional, Iterable, Mapping, Tuple, Dict, List, Sequence


class AvailableModulesTreeWidget(QtWidgets.QTreeWidget):
    """
    """
    def __init__(self,
                 modules: Optional[Iterable[str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self.setColumnCount(2)
        self.setHeaderLabels(('Base', 'module.Class'))
        self.setSelectionMode(self.ExtendedSelection)
        self.setEditTriggers(self.EditTrigger.NoEditTriggers)
        self.setDragEnabled(True)

        self.top_level_items = dict()
        self.clear_modules()
        if modules is not None:
            self.set_modules(sorted(modules))

    @property
    def modules(self) -> List[str]:
        modules = list()
        for base, top_item in self.top_level_items.items():
            items = [top_item.child(index) for index in range(top_item.childCount())]
            modules.extend(f'{base}.{it.text(1)}' for it in items)
        return modules

    def set_modules(self, modules: Iterable[str]) -> None:
        # Clear all modules
        self._clear_modules()
        # Add new modules
        for module in modules:
            self._add_module(module)
        # Resize columns
        self.resize_columns_to_content()

    def add_module(self, module: str) -> None:
        self._add_module(module)
        self.resize_columns_to_content()

    def _add_module(self, module: str) -> None:
        base, module_class = module.split('.', 1)
        item = QtWidgets.QTreeWidgetItem()
        item.setText(1, module_class)
        item.setFlags(
            QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsDragEnabled
        )
        self.top_level_items[base].addChild(item)

    def remove_module(self, module: str) -> None:
        self._remove_module(module)
        self.resize_columns_to_content()

    def _remove_module(self, module: str) -> None:
        base, module_class = module.split('.', 1)
        top_level_item = self.top_level_items[base]
        for index in range(top_level_item.childCount()):
            child = top_level_item.child(index)
            if child.text(1) == module_class:
                top_level_item.removeChild(child)
                break

    def clear_modules(self) -> None:
        self._clear_modules()
        self.resize_columns_to_content()

    def _clear_modules(self) -> None:
        self.clear()
        for disp_base in ('GUI', 'Logic', 'Hardware'):
            item = QtWidgets.QTreeWidgetItem()
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            item.setText(0, disp_base)
            self.addTopLevelItem(item)
            item.setExpanded(True)
            self.top_level_items[disp_base.lower()] = item

    def resize_columns_to_content(self) -> None:
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)

    def mimeData(self, items: Sequence) -> QtCore.QMimeData:
        """ Add text to mime data. This is the quick (but not necessarily dirty) way.
        """
        texts = tuple(f'{it.parent().text(0).lower()}.{it.text(1)}' for it in items)
        mime = super().mimeData(items)
        mime.setText(';'.join(texts))
        return mime


class ConfigModulesTreeWidget(QtWidgets.QTreeWidget):
    """
    """

    def __init__(self,
                 named_modules: Optional[Mapping[str, str]] = None,
                 unnamed_modules: Optional[Iterable[str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self.setDragEnabled(False)
        self.setDropIndicatorShown(False)
        self.setAcceptDrops(True)
        self.setColumnCount(3)
        self.setHeaderLabels(['Base', 'Name', 'module.Class'])
        self.setSelectionMode(self.SingleSelection)
        self.setEditTriggers(self.EditTrigger.NoEditTriggers)
        self.itemDoubleClicked.connect(self._edit_item_column)

        self.top_level_items = dict()
        self.set_modules(named_modules, unnamed_modules)
        self._valid_foreground = next(iter(self.top_level_items.values())).foreground(0)
        self._name_regex = compile(r'^[a-zA-Z_]+[a-zA-Z0-9_]*$')

    @property
    def modules(self) -> Tuple[Dict[str, str], List[str]]:
        named_modules = dict()
        unnamed_modules = list()
        for base, top_item in self.top_level_items.items():
            items = [top_item.child(index) for index in range(top_item.childCount())]
            for it in items:
                name = it.text(1)
                module = f'{base}.{it.text(2)}'
                if self.is_valid_name(name) and name not in named_modules:
                    named_modules[name] = module
                else:
                    unnamed_modules.append(module)
        return named_modules, unnamed_modules

    def set_modules(self,
                    named_modules: Optional[Mapping[str, str]] = None,
                    unnamed_modules: Optional[Iterable[str]] = None
                    ) -> None:
        if named_modules is None:
            named_modules = dict()
        if unnamed_modules is None:
            unnamed_modules = list()
        # Clear all modules
        self._clear_modules()
        # Add new modules
        for name, module in named_modules.items():
            self._add_module(module, name)
        for module in unnamed_modules:
            self._add_module(module)
        # Resize columns
        self.resize_columns_to_content()

    def add_module(self, module: str, name: Optional[str] = None) -> None:
        self._add_module(module, name)
        self.resize_columns_to_content()

    def _add_module(self, module: str, name: Optional[str] = None) -> None:
        base, module_class = module.split('.', 1)
        item = QtWidgets.QTreeWidgetItem()
        item.setText(1, '<enter unique name>' if name is None else name)
        item.setText(2, module_class)
        item.setFlags(
            QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable
        )
        self.top_level_items[base].addChild(item)

    def remove_module(self, name: str) -> None:
        self._remove_module(name)
        self.resize_columns_to_content()

    def _remove_module(self, name: str) -> None:
        found = False
        for base, top_item in self.top_level_items.items():
            for index in range(top_item.childCount()):
                child = top_item.child(index)
                if child.text(1) == name:
                    top_item.removeChild(child)
                    found = True
                    break
            if found:
                break

    def clear_modules(self) -> None:
        self._clear_modules()
        self.resize_columns_to_content()

    def _clear_modules(self) -> None:
        self.clear()
        for disp_base in ['GUI', 'Logic', 'Hardware']:
            item = QtWidgets.QTreeWidgetItem()
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            item.setText(0, disp_base)
            self.addTopLevelItem(item)
            item.setExpanded(True)
            self.top_level_items[disp_base.lower()] = item

    def resize_columns_to_content(self) -> None:
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)

    def is_valid_name(self, name: str) -> bool:
        set_names = list()
        for base, top_item in self.top_level_items.items():
            items = [top_item.child(index) for index in range(top_item.childCount())]
            set_names.extend(it.text(1) for it in items)
        return self._name_regex.match(name) is not None and set_names.count(name) <= 1

    @QtCore.Slot(QtWidgets.QTreeWidgetItem, int)
    def _edit_item_column(self, item: QtWidgets.QTreeWidgetItem, column: int) -> None:
        if item and column == 1 and item.parent() is not None:
            self.editItem(item, column)


class SelectedModulesTreeWidget(ConfigModulesTreeWidget):
    """
    """
    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        if isinstance(event.source(), AvailableModulesTreeWidget):
            full_text = event.mimeData().text()
            for module in full_text.split(';'):
                self._add_module(module)
            self.resize_columns_to_content()
            event.accept()
        else:
            event.ignore()

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key_Delete:
            for item in self.selectedItems():
                if item.parent() is not None:
                    item.parent().removeChild(item)
            event.accept()
        else:
            super().keyPressEvent(event)



# -*- coding: utf-8 -*-

"""
JSON schema to be used by jsonschema.validate on YAML qudi configuration files.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['config_schema', 'local_module_config_schema', 'remote_module_config_schema']

from typing import Dict, Any


__module_name_pattern = r'[a-zA-Z_]+[a-zA-Z0-9_]*'


def config_schema() -> Dict[str, Any]:
    """Creates and returns the JSON schema for a qudi configuration."""
    return {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'global': {
                'type': 'object',
                'additionalProperties': True,
                'default': dict(),
                'properties': {
                    'startup_modules': {
                        'type': 'array',
                        'uniqueItems': True,
                        'items': {
                            'type': 'string',
                            'pattern': f'^{__module_name_pattern}$'
                        },
                        'default': list()
                    },
                    'remote_modules_server': {
                        'type': ['null', 'object'],
                        'required': ['address', 'port'],
                        'default': None,
                        'additionalProperties': False,
                        'properties': {
                            'address': {
                                'type': 'string',
                            },
                            'port': {
                                'type': 'integer',
                                'minimum': 0,
                                'maximum': 65535
                            },
                            'certfile': {
                                'type': ['null', 'string'],
                                'default': None
                            },
                            'keyfile': {
                                'type': ['null', 'string'],
                                'default': None
                            }
                        }
                    },
                    'namespace_server_port': {
                        'type': 'integer',
                        'minimum': 0,
                        'maximum': 65535,
                        'default': 18861
                    },
                    'force_remote_calls_by_value': {
                        'type': 'boolean',
                        'default': True
                    },
                    'hide_manager_window': {
                        'type': 'boolean',
                        'default': False
                    },
                    'stylesheet': {
                        'type': 'string',
                        'pattern': r'^[^.]+\.qss$',
                        'default': 'qdark.qss'
                    },
                    'daily_data_dirs': {
                        'type': 'boolean',
                        'default': True
                    },
                    'default_data_dir': {
                        'type': ['null', 'string'],
                        'default': None
                    },
                    'extension_paths': {
                        'type': 'array',
                        'uniqueItems': True,
                        'items': {
                            'type': 'string'
                        },
                        'default': list()
                    }
                }
            },
            'gui': {
                'type': 'object',
                'propertyNames': {
                    'pattern': f'^{__module_name_pattern}$'
                },
                'additionalProperties': {
                    '$ref': '#/$defs/local_module'
                },
                'default': dict()
            },
            'logic': {
                'type': 'object',
                'propertyNames': {
                    'pattern': f'^{__module_name_pattern}$'
                },
                'additionalProperties': {
                    'oneOf': [
                        {'$ref': '#/$defs/local_module'},
                        {'$ref': '#/$defs/remote_module'}
                    ]
                },
                'default': dict()
            },
            'hardware': {
                'type': 'object',
                'propertyNames': {
                    'pattern': f'^{__module_name_pattern}$'
                },
                'additionalProperties': {
                    'oneOf': [
                        {'$ref': '#/$defs/local_module'},
                        {'$ref': '#/$defs/remote_module'}
                    ]
                },
                'default': dict()
            }
        },

        '$defs': {
            'local_module': local_module_config_schema(),
            'remote_module': remote_module_config_schema()
        }
    }


def local_module_config_schema() -> Dict[str, Any]:
    """Creates and returns the JSON schema for a single qudi local module configuration."""
    return {
        'type': 'object',
        'required': ['module.Class'],
        'additionalProperties': False,
        'properties': {
            'module.Class': {
                'type': 'string',
                'pattern': f'^{__module_name_pattern}(\\.{__module_name_pattern})*$',
            },
            'allow_remote': {
                'type': 'boolean',
                'default': False
            },
            'connect': {
                'type': 'object',
                'additionalProperties': {
                    'type': 'string',
                    'pattern': f'^{__module_name_pattern}$'
                },
                'default': dict()
            },
            'options': {
                'type': 'object',
                'additionalProperties': True,
                'default': dict()
            }
        }
    }


def remote_module_config_schema() -> Dict[str, Any]:
    """Creates and returns the JSON schema for a single qudi remote module configuration."""
    return {
        'type': 'object',
        'required': ['native_module_name', 'address', 'port'],
        'additionalProperties': False,
        'properties': {
            'native_module_name': {
                'type': 'string',
                'pattern': f'^{__module_name_pattern}$'
            },
            'address': {
                'type': 'string',
            },
            'port': {
                'type': 'integer',
                'minimum': 0,
                'maximum': 65535
            },
            'certfile': {
                'type': ['null', 'string'],
                'default': None
            },
            'keyfile': {
                'type': ['null', 'string'],
                'default': None
            }
        }
    }



# -*- coding: utf-8 -*-
"""
This file contains the qudi log widget class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtCore, QtGui, QtWidgets
from qudi.core.logger import get_record_table_model


class LogFilterProxy(QtCore.QSortFilterProxyModel):
    """
    A subclass of QProxyFilterModel that determines which log entries contained in the log model
    are shown in the view.
    """

    def __init__(self, parent=None):
        """
        Create the LogFilterProxy.

        Parameters
        ----------
        parent : QObject
            Parent object of the filter.
        """
        super().__init__(parent)
        self._show_levels = frozenset({'debug', 'info', 'warning', 'error', 'critical'})

    def filterAcceptsRow(self, source_row, source_parent):
        """
        Determine whether a row (log entry) should be shown.

        Parameters
        ----------
        source_row : QModelIndex
            The row in the source model that needs to be filtered.
        source_parent : QModelIndex
            The parent model index.

        Returns
        -------
        bool
            `True` if the row (log entry) should be shown, `False` otherwise.
        """
        model = self.sourceModel()
        level = model.data(model.index(source_row, 1), QtCore.Qt.DisplayRole)
        if level is None:
            return False
        return level in self._show_levels

    def set_levels(self, levels):
        """
        Set which types of messages are shown through the filter.

        Parameters
        ----------
        levels : set of str
            Set of all message levels that should be shown.
        """
        self._show_levels = frozenset(levels)
        self.invalidateFilter()


class SelectableTextDelegate(QtWidgets.QStyledItemDelegate):
    """A subclass of QStyledItemDelegate to display a text editor for copying text fragments.
    """
    def createEditor(self, parent, option, index):
        """
        Overwrite method from base class QStyledItemDelegate to show a read-only QLabel widget.
        This is necessary to disable editing by the user but still allow text to be marked and copied.

        Parameters
        ----------
        parent : QObject
            The parent object for the editor to be created.
        option : QStyleOptionViewItem
            Display options for the editor widget.
        index : QModelIndex
            Data model index.

        Returns
        -------
        QLabel
            Instance of QLabel.
        """
        editor = QtWidgets.QLabel(parent)
        editor.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        editor.setAlignment(option.displayAlignment)
        return editor

    def setEditorData(self, editor, index):
        """
        Overwrite method from the base class QStyledItemDelegate to fill the QLineEdit widget with data.

        Parameters
        ----------
        editor : QLineEdit
            Editor widget to be populated with data.
        index : QModelIndex
            Data model index.
        """
        data = index.data(QtCore.Qt.EditRole)
        editor.setText(f' {data}')


class LogTableWidget(QtWidgets.QTableView):
    """Customized QTableView including the model for display of logging entries.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        record_model = get_record_table_model()
        if record_model is None:
            raise RuntimeError('No record table model set up in qudi.core.logger')
        self.filter_model = LogFilterProxy()
        self.filter_model.setSourceModel(record_model)
        self.setModel(self.filter_model)

        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        self.setEditTriggers(QtWidgets.QTableView.DoubleClicked)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QtWidgets.QTableView.NoSelection)
        self.setHorizontalScrollMode(QtWidgets.QTableView.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QTableView.ScrollPerPixel)
        self.setShowGrid(False)
        self.setCornerButtonEnabled(False)
        self.horizontalHeader().setCascadingSectionResizes(True)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().hide()
        self.setItemDelegate(SelectableTextDelegate())
        self.horizontalHeader().setMinimumSectionSize(50)
        self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)
        self.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.Fixed)
        # Set fixed with for "Time" and "Level" columns since they contain fixed width strings
        metrics = QtGui.QFontMetrics(self.font())
        self.setColumnWidth(0, metrics.horizontalAdvance(' 5555-55-55 55:55:55 '))
        self.setColumnWidth(1, metrics.horizontalAdvance(' warning '))
        # Estimate starting width of "Source" column
        self.setColumnWidth(2, metrics.horizontalAdvance(__name__ * 2))

        self.filter_model.rowsInserted.connect(self._entry_added)

    @QtCore.Slot(QtCore.QModelIndex, int, int)
    def _entry_added(self, parent, first, last):
        while first <= last:
            self.resizeRowToContents(first)
            first += 1
        self.scrollToBottom()

    def set_level_filter(self, show_levels):
        self.filter_model.set_levels(show_levels)
        self.scrollToBottom()


class LogWidget(QtWidgets.QSplitter):
    """A widget to show log entries and filter them.
    """

    def __init__(self, parent=None, debug_mode=False, **kwargs):
        """
        Creates the log widget.

        Parameters
        ----------
        parent : QObject
            Qt parent object for the log widget.
        manager : Manager
            Manager instance this widget belongs to.
        """
        super().__init__(QtCore.Qt.Horizontal, parent, **kwargs)

        # Build GUI elements
        # Set up QTableView to display log entries
        self.log_tablewidget = LogTableWidget()
        self.log_tablewidget.setObjectName('log_table_widget')

        # Set up QTreeWidget for log filter ui
        self.filter_treewidget = QtWidgets.QTreeWidget()
        self.filter_treewidget.setObjectName('filter_treewidget')
        self.filter_treewidget.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                             QtWidgets.QSizePolicy.Preferred)
        self.filter_treewidget.setMinimumSize(210, 0)
        self.filter_treewidget.setEditTriggers(QtWidgets.QTreeWidget.NoEditTriggers)
        self.filter_treewidget.setDropIndicatorShown(False)
        self.filter_treewidget.setDragEnabled(False)
        self.filter_treewidget.setSelectionMode(QtWidgets.QTreeWidget.NoSelection)
        self.filter_treewidget.setSelectionBehavior(QtWidgets.QTreeWidget.SelectItems)
        self.filter_treewidget.setColumnCount(1)
        self.filter_treewidget.setHeaderLabels(('Display:',))
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, 'All message types:')
        item.setCheckState(0, QtCore.Qt.Checked)
        log_levels = ('debug', 'info', 'warning', 'error', 'critical')[int(not debug_mode):]
        for text in log_levels:
            child_item = QtWidgets.QTreeWidgetItem()
            child_item.setText(0, text)
            child_item.setCheckState(0, QtCore.Qt.Checked)
            item.addChild(child_item)
        self.filter_treewidget.addTopLevelItem(item)
        self.filter_treewidget.expandItem(item)
        self.log_tablewidget.set_level_filter(log_levels)

        # embed log view and filter tree into QSplitter widget
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        self.addWidget(self.log_tablewidget)
        self.addWidget(self.filter_treewidget)
        self.setStretchFactor(0, 1)

        # connect signals
        self.filter_treewidget.itemChanged.connect(self.update_filter_state)

    @QtCore.Slot(object, int)
    def update_filter_state(self, item, column):
        """
        Update the log view based on the filter widget check states and synchronize check box states.

        Parameters
        ----------
        item : int
            Item number.
        column : int
            Column number.
        """
        # check all / uncheck all state
        show_all_item = self.filter_treewidget.topLevelItem(0)
        level_items = [show_all_item.child(ii) for ii in range(show_all_item.childCount())]
        if item is show_all_item:
            self.filter_treewidget.expandItem(item)
            if show_all_item.checkState(0):
                self.filter_treewidget.blockSignals(True)
                for it in level_items:
                    it.setCheckState(0, QtCore.Qt.Checked)
                self.filter_treewidget.blockSignals(False)
            else:
                # Prevent user from unchecking "show all"
                self.filter_treewidget.blockSignals(True)
                show_all_item.setCheckState(0, QtCore.Qt.Checked)
                self.filter_treewidget.blockSignals(False)
        else:
            show_all = all(it.checkState(0) for it in level_items)
            self.filter_treewidget.blockSignals(True)
            show_all_item.setCheckState(0, QtCore.Qt.Checked if show_all else QtCore.Qt.Unchecked)
            self.filter_treewidget.blockSignals(False)

        # set level filters
        level_filter = {str(it.text(0)) for it in level_items if it.checkState(0)}
        self.log_tablewidget.set_level_filter(level_filter)



# -*- coding: utf-8 -*-

"""
Convenience classes to get vertical and horizontal separator lines in a QLayout as one-liner.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['VerticalLine', 'HorizontalLine']

from PySide2 import QtWidgets
from typing import Optional


class VerticalLine(QtWidgets.QFrame):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)


class HorizontalLine(QtWidgets.QFrame):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)



# -*- coding: utf-8 -*-

"""
This file contains models of Poissonian fitting routines for qudi based on the lmfit package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('DoublePoissonian', 'Poissonian', 'multiple_poissonian')

import numpy as np
from scipy.special import gammaln, xlogy
from qudi.util.fit_models.model import FitModelBase, estimator
from qudi.util.fit_models.helpers import smooth_data, sort_check_data, estimate_double_peaks
from qudi.util.fit_models.gaussian import multiple_gaussian


def multiple_poissonian(x, mus, amplitudes):
    """ Mathematical definition of the sum of multiple scaled Poissonian distributions without any
    bias.

    WARNING: Iterable parameters "mus" and "amplitudes" must have the same length.

    Parameters
    ----------
    x : float
        The independent variable to calculate the Poissonian.
    mus : iterable
        Iterable containing center positions (means) for all Poissonians.
    amplitudes : iterable
        Iterable containing amplitudes for all Poissonians.

    Returns
    -------
    float
        The result given x for the sum of scaled Poissonian distributions.

    """
    assert len(mus) == len(amplitudes)

    # Use the numerically more stable definition of the Poisson probability mass function.
    # The standard definition is quickly blowing up for increasing values of x.
    #
    # Due to the fact that the poissonian distribution is approaching a normal distribution for
    # large values of mu, we define a cut-off value of 1e6. If our independent variables x are at
    # or above this value, we will switch to calculating a normal distribution. This ensures that
    # this function will remain numerically stable for very large values of x and mu.
    if min(x) < 1e6:
        return sum(np.exp(xlogy(x, mu) - gammaln(x + 1) - mu) for mu, amp in
                   zip(mus, amplitudes, amplitudes))
    else:
        return multiple_gaussian(x, mus, np.sqrt(mus), amplitudes)


class Poissonian(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=1., min=0, max=np.inf)
        self.set_param_hint('mu', value=1., min=0, max=np.inf)

    @staticmethod
    def _model_function(x, offset, mu, amplitude):
        return offset + multiple_poissonian(x, (mu,), (amplitude,))

    @estimator('default')
    def estimate(self, data, x):
        data, x = sort_check_data(data, x)
        filter_width = max(1, int(round(len(x) / 20)))
        data_smoothed, _ = smooth_data(data, filter_width)

        # estimate offset and level data
        offset = min(data_smoothed)
        data_smoothed -= offset

        # estimate other parameters
        mu = x[np.argmax(data_smoothed)]
        amplitude = max(data_smoothed) / self._model_function(mu, 0, mu, 1)

        x_spacing = min(abs(np.ediff1d(x)))
        x_span = abs(x[-1] - x[0])
        data_span = abs(max(data) - min(data))

        estimate = self.make_params()
        estimate['mu'].set(value=mu,
                           min=max(x_spacing, min(x) - x_span / 2),
                           max=min(x_span, max(x) + x_span / 2))
        estimate['amplitude'].set(value=amplitude, min=0, max=2 * amplitude)
        estimate['offset'].set(value=offset,
                               min=min(data) - data_span / 2,
                               max=max(data) + data_span / 2)
        return estimate

    @estimator('No Offset')
    def estimate_no_offset(self, data, x):
        estimate = self.estimate(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate


class DoublePoissonian(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=1., min=0, max=np.inf)
        self.set_param_hint('amplitude_2', value=1., min=0, max=np.inf)
        self.set_param_hint('mu_1', value=1., min=0, max=np.inf)
        self.set_param_hint('mu_2', value=2., min=0, max=np.inf)

    @staticmethod
    def _model_function(x, offset, mu_1, mu_2, amplitude_1, amplitude_2):
        return offset + multiple_poissonian(x, (mu_1, mu_2), (amplitude_1, amplitude_2))

    @estimator('default')
    def estimate(self, data, x):
        data, x = sort_check_data(data, x)
        data_smoothed, filter_width = smooth_data(data)
        # estimate offset and level data
        offset = min(data_smoothed)
        data_smoothed -= offset

        estimate, limits = estimate_double_peaks(data_smoothed, x, filter_width)

        params = self.make_params()
        params['amplitude_1'].set(value=estimate['height'][0],
                                  min=limits['height'][0][0],
                                  max=limits['height'][0][1])
        params['amplitude_2'].set(value=estimate['height'][1],
                                  min=limits['height'][1][0],
                                  max=limits['height'][1][1])
        params['center_1'].set(value=estimate['center'][0],
                               min=limits['center'][0][0],
                               max=limits['center'][0][1])
        params['center_2'].set(value=estimate['center'][1],
                               min=limits['center'][1][0],
                               max=limits['center'][1][1])
        params['sigma_1'].set(value=estimate['fwhm'][0] / 2.3548,
                              min=limits['fwhm'][0][0] / 2.3548,
                              max=limits['fwhm'][0][1] / 2.3548)
        params['sigma_2'].set(value=estimate['fwhm'][1] / 2.3548,
                              min=limits['fwhm'][1][0] / 2.3548,
                              max=limits['fwhm'][1][1] / 2.3548)
        return params

    @estimator('No Offset')
    def estimate_no_offset(self, data, x):
        estimate = self.estimate(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate



# -*- coding: utf-8 -*-

"""
This file contains customized pyqtgraph graphics items to be used as data marker in 1D and 2D plots.
This is an attempt to provide a somewhat unified interface to these markers contrary to the
heterogeneous interfaces of pyqtgraph items.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['InfiniteCrosshair', 'InfiniteLine', 'LinearRegion', 'Rectangle',
           'InfiniteCrosshairRectangle']

from math import isinf
from typing import Union, Tuple, Optional, List, Sequence, Any

from PySide2 import QtCore
from pyqtgraph import ViewBox
from pyqtgraph import ROI as _ROI
from pyqtgraph import LinearRegionItem as _LinearRegionItem
from pyqtgraph import InfiniteLine as _InfiniteLine
from qudi.util.widgets.plotting.roi import RectangleROI as _RectangleROI


class InfiniteCrosshair(QtCore.QObject):
    """ Represents a crosshair (two perpendicular infinite lines) """

    _default_pen = {'color': '#00ff00', 'width': 1}
    _default_hover_pen = {'color': '#ffff00', 'width': 1}

    sigPositionChanged = QtCore.Signal(tuple)  # current_pos
    # start_pos, current_pos, is_start, is_finished
    sigPositionDragged = QtCore.Signal(tuple, tuple, bool, bool)

    def __init__(self,
                 viewbox: ViewBox,
                 position: Optional[Tuple[float, float]] = (0, 0),
                 bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 movable: Optional[bool] = True,
                 pen: Optional[Any] = None,
                 hover_pen: Optional[Any] = None,
                 ) -> None:
        super().__init__(parent=viewbox)
        if position is None:
            position = (0, 0)
        if movable is None:
            movable = True
        if pen is None:
            pen = self._default_pen
        if hover_pen is None:
            hover_pen = self._default_hover_pen

        self._bounds = self._normalize_bounds(bounds)
        self._z_value = None
        self._is_dragged = False
        self.vline = _InfiniteLine(pos=position[0],
                                   bounds=self._bounds[0],
                                   angle=90,
                                   movable=movable,
                                   pen=pen,
                                   hoverPen=hover_pen)
        self.hline = _InfiniteLine(pos=position[1],
                                   bounds=self._bounds[1],
                                   angle=0,
                                   movable=movable,
                                   pen=pen,
                                   hoverPen=hover_pen)

        self.vline.sigDragged.connect(self._line_dragged)
        self.vline.sigPositionChanged.connect(self._line_changed)
        self.vline.sigPositionChangeFinished.connect(self._line_drag_finished)
        self.hline.sigDragged.connect(self._line_dragged)
        self.hline.sigPositionChanged.connect(self._line_changed)
        self.hline.sigPositionChangeFinished.connect(self._line_drag_finished)
        self.show()

    @property
    def movable(self) -> bool:
        return bool(self.vline.movable)

    def set_movable(self, movable: bool) -> None:
        """ (Un-)Set the crosshair movable (draggable by mouse cursor) """
        self.vline.setMovable(movable)
        self.hline.setMovable(movable)

    @property
    def z_value(self) -> Union[None, int]:
        return self._z_value

    def set_z_value(self, value: int) -> None:
        self.vline.setZValue(value)
        self.hline.setZValue(value)
        self._z_value = value

    @property
    def position(self) -> Tuple[float, float]:
        return self.vline.value(), self.hline.value()

    def set_position(self, pos: Tuple[float, float]) -> None:
        self.vline.blockSignals(True)
        self.hline.blockSignals(True)
        try:
            self.vline.setPos(pos[0])
            self.hline.setPos(pos[1])
        finally:
            self.vline.blockSignals(False)
            self.hline.blockSignals(False)
        self.sigPositionChanged.emit(self.position)

    @property
    def bounds(self) -> List[Tuple[Union[None, float], Union[None, float]]]:
        return self._bounds.copy()

    def set_bounds(self,
                   bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
                   ) -> None:
        """ Sets a range boundary for the crosshair position """
        self._bounds = self._normalize_bounds(bounds)
        self.vline.setBounds(self._bounds[0])
        self.hline.setBounds(self._bounds[1])

    def show(self):
        view = self.parent()
        if self.vline not in view.addedItems:
            view.addItem(self.vline)
            if self._z_value is not None:
                self.vline.setZValue(self._z_value)
        if self.hline not in view.addedItems:
            view.addItem(self.hline)
            if self._z_value is not None:
                self.hline.setZValue(self._z_value)

    def hide(self):
        view = self.parent()
        if self.vline in view.addedItems:
            view.removeItem(self.vline)
        if self.hline in view.addedItems:
            view.removeItem(self.hline)

    def set_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the crosshair lines.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        self.vline.setPen(pen)
        self.hline.setPen(pen)

    def set_hover_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the crosshair lines when the mouse cursor is
        hovering over them.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        self.vline.setHoverPen(pen)
        self.hline.setHoverPen(pen)

    @staticmethod
    def _normalize_bounds(
            bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
            ) -> List[Tuple[Union[None, float], Union[None, float]]]:
        if bounds is None:
            bounds = [(None, None), (None, None)]
        else:
            bounds = [list(span) for span in bounds]
            # Replace inf values by None
            for span in bounds:
                for ii, val in enumerate(span):
                    try:
                        if isinf(val):
                            span[ii] = None
                    except TypeError:
                        pass
            # Sort spans in ascending order
            try:
                bounds[0] = tuple(sorted(bounds[0]))
            except TypeError:
                bounds[0] = tuple(bounds[0])
            try:
                bounds[1] = tuple(sorted(bounds[1]))
            except TypeError:
                bounds[1] = tuple(bounds[1])
        return bounds

    def _line_dragged(self, line: Optional[_InfiniteLine] = None) -> None:
        if self._is_dragged:
            is_start = False
        else:
            self._is_dragged = True
            is_start = True

        current_pos = self.position
        if line is self.vline:
            start_pos = (self.vline.startPosition[0], current_pos[1])
        else:
            start_pos = (current_pos[0], self.hline.startPosition[1])

        if line.moving:
            is_finished = False
        else:
            self._is_dragged = False
            is_finished = True

        self.sigPositionDragged.emit(start_pos, current_pos, is_start, is_finished)

    def _line_drag_finished(self, line: Optional[_InfiniteLine] = None) -> None:
        self._is_dragged = False
        current_pos = self.position
        if line is self.vline:
            start_pos = (self.vline.startPosition[0], current_pos[1])
        else:
            start_pos = (current_pos[0], self.hline.startPosition[1])
        self.sigPositionDragged.emit(start_pos, current_pos, False, True)

    def _line_changed(self, line: Optional[_InfiniteLine] = None) -> None:
        self.sigPositionChanged.emit(self.position)


class InfiniteLine(QtCore.QObject):
    """ Represents a horizontal or vertical infinite line data marker """

    _default_pen = {'color': '#00ff00', 'width': 1}
    _default_hover_pen = {'color': '#ffff00', 'width': 1}

    sigPositionChanged = QtCore.Signal(object)  # current_pos
    # start_pos, current_pos, is_start, is_finished
    sigPositionDragged = QtCore.Signal(object, object, bool, bool)

    def __init__(self,
                 viewbox: ViewBox,
                 orientation: QtCore.Qt.Orientation,
                 position: Optional[float] = 0,
                 bounds: Optional[Tuple[Union[None, float]]] = None,
                 movable: Optional[bool] = True,
                 pen: Optional[Any] = None,
                 hover_pen: Optional[Any] = None,
                 ) -> None:
        super().__init__(parent=viewbox)
        if position is None:
            position = 0
        if movable is None:
            movable = True
        if pen is None:
            pen = self._default_pen
        if hover_pen is None:
            hover_pen = self._default_hover_pen

        self._bounds = self._normalize_bounds(bounds)
        self._z_value = None
        self._is_dragged = False
        self.line = _InfiniteLine(pos=position,
                                  angle=0 if orientation is QtCore.Qt.Horizontal else 90,
                                  movable=movable,
                                  pen=pen,
                                  hoverPen=hover_pen)

        self.line.sigDragged.connect(self._line_dragged)
        self.line.sigPositionChanged.connect(self._line_changed)
        self.line.sigPositionChangeFinished.connect(self._line_drag_finished)

        self.show()

    @property
    def orientation(self) -> QtCore.Qt.Orientation:
        return QtCore.Qt.Vertical if self.line.angle == 90 else QtCore.Qt.Horizontal

    @property
    def movable(self) -> bool:
        return bool(self.line.movable)

    def set_movable(self, movable: bool) -> None:
        self.line.setMovable(movable)

    @property
    def z_value(self) -> Union[None, int]:
        return self._z_value

    def set_z_value(self, value: int) -> None:
        self.line.setZValue(value)
        self._z_value = value

    @property
    def position(self) -> float:
        return self.line.value()

    def set_position(self, position: float) -> None:
        self.line.blockSignals(True)
        try:
            self.line.setPos(position)
        finally:
            self.line.blockSignals(False)
        self.sigPositionChanged.emit(self.position)

    @property
    def bounds(self) -> Tuple[Union[None, float], Union[None, float]]:
        return self._bounds

    def set_bounds(self,
                   bounds: Union[None, Tuple[Union[None, float], Union[None, float]]]
                   ) -> None:
        """ Sets a range boundary for the line position """
        self._bounds = self._normalize_bounds(bounds)
        self.line.setBounds(self._bounds)

    def show(self):
        view = self.parent()
        if self.line not in view.addedItems:
            view.addItem(self.line)
            if self._z_value is not None:
                self.line.setZValue(self._z_value)

    def hide(self):
        view = self.parent()
        if self.line in view.addedItems:
            view.removeItem(self.line)

    def set_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the line.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        self.line.setPen(pen)

    def set_hover_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the line when the mouse cursor is hovering over it.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        self.line.setHoverPen(pen)

    @staticmethod
    def _normalize_bounds(bounds: Union[None, Tuple[Union[None, float], Union[None, float]]]
                          ) -> Tuple[Union[None, float], Union[None, float]]:
        if bounds is None:
            bounds = (None, None)
        else:
            # Replace inf values by None
            try:
                if isinf(bounds[0]):
                    bounds = (None, bounds[1])
            except TypeError:
                pass
            try:
                if isinf(bounds[1]):
                    bounds = (bounds[0], None)
            except TypeError:
                pass
            # Sort spans in ascending order
            try:
                bounds = tuple(sorted(bounds))
            except TypeError:
                bounds = tuple(bounds)
        return bounds

    def _line_dragged(self, line: Optional[_InfiniteLine] = None) -> None:
        if self._is_dragged:
            is_start = False
        else:
            self._is_dragged = True
            is_start = True

        start_pos = self.line.startPosition[0 if self.line.angle == 90 else 1]
        self.sigPositionDragged.emit(start_pos, self.position, is_start, False)

    def _line_drag_finished(self, line: Optional[_InfiniteLine] = None) -> None:
        self._is_dragged = False
        start_pos = self.line.startPosition[0 if self.line.angle == 90 else 1]
        self.sigPositionDragged.emit(start_pos, self.position, False, True)

    def _line_changed(self, line: Optional[_InfiniteLine] = None) -> None:
        self.sigPositionChanged.emit(self.position)


class LinearRegion(QtCore.QObject):
    """
    """
    _default_pen = {'color': '#00ff00', 'width': 1}
    _default_hover_pen = {'color': '#ffff00', 'width': 1}
    _default_brush = None
    _default_hover_brush = None

    sigAreaChanged = QtCore.Signal(tuple)  # current_area
    # start_area, current_area, is_start, is_finished
    sigAreaDragged = QtCore.Signal(tuple, tuple, bool, bool)

    def __init__(self,
                 viewbox: ViewBox,
                 orientation: QtCore.Qt.Orientation,
                 span: Optional[Tuple[float, float]] = (0, 1),
                 bounds: Optional[Tuple[Union[None, float], Union[None, float]]] = None,
                 movable: Optional[bool] = True,
                 pen: Optional[Any] = None,
                 hover_pen: Optional[Any] = None,
                 brush: Optional[Any] = None,
                 hover_brush: Optional[Any] = None
                 ) -> None:
        super().__init__(parent=viewbox)
        if span is None:
            span = (0, 1)
        if movable is None:
            movable = True
        if pen is None:
            pen = self._default_pen
        if hover_pen is None:
            hover_pen = self._default_hover_pen
        if brush is None:
            brush = self._default_brush
        if hover_brush is None:
            hover_brush = self._default_hover_brush
        orientation = 'vertical' if orientation == QtCore.Qt.Vertical else 'horizontal'

        self._bounds = self._normalize_bounds(bounds)
        self.region = _LinearRegionItem(values=span,
                                        orientation=orientation,
                                        brush=brush,
                                        pen=pen,
                                        hoverBrush=hover_brush,
                                        hoverPen=hover_pen,
                                        movable=movable,
                                        bounds=self._bounds,
                                        swapMode='sort')

        self._z_value = None
        self._is_dragged = False
        self._start_area = self.area

        self.region.sigRegionChangeFinished.connect(self._region_change_finished)
        self.region.sigRegionChanged.connect(self._region_changed)

        self.show()

    @property
    def orientation(self) -> QtCore.Qt.Orientation:
        return QtCore.Qt.Vertical if self.region.orientation == 'vertical' else QtCore.Qt.Horizontal

    @property
    def movable(self) -> bool:
        return self.region.movable

    def set_movable(self, movable: bool) -> None:
        self.region.setMovable(movable)

    @property
    def z_value(self) -> Union[None, int]:
        return self._z_value

    def set_z_value(self, value: int) -> None:
        self.region.setZValue(value)
        self._z_value = value

    @property
    def area(self) -> Tuple[float, float]:
        return self.region.getRegion()

    def set_area(self, area: Tuple[float, float]) -> None:
        self.region.blockSignals(True)
        try:
            self.region.setRegion(area)
        finally:
            self.region.blockSignals(False)
        self.sigAreaChanged.emit(self.area)

    @property
    def bounds(self) -> Tuple[Union[None, float], Union[None, float]]:
        return self._bounds

    def set_bounds(self,
                   bounds: Union[None, Tuple[Union[None, float], Union[None, float]]]
                   ) -> None:
        self._bounds = self._normalize_bounds(bounds)
        self.region.setBounds(self._bounds)

    def show(self):
        view = self.parent()
        if self.region not in view.addedItems:
            view.addItem(self.region)
            if self._z_value is not None:
                self.region.setZValue(self._z_value)

    def hide(self):
        view = self.parent()
        if self.region in view.addedItems:
            view.removeItem(self.region)

    def set_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the lines.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        for line in self.region.lines:
            line.setPen(pen)

    def set_hover_pen(self, pen: Any) -> None:
        """ Sets the pen to be used for drawing the lines when the mouse cursor is hovering over it.
        Given parameter must be compatible with pyqtgraph.mkPen()
        """
        for line in self.region.lines:
            line.setHoverPen(pen)

    def set_brush(self, brush: Any) -> None:
        """ Sets the brush to be used for filling the area between the lines.
        Given parameter must be compatible with pyqtgraph.mkBrush()
        """
        self.region.setBrush(brush)

    def set_hover_brush(self, brush: Any) -> None:
        """ Sets the brush to be used for filling the area between the lines when the mouse cursor
        is hovering over it.
        Given parameter must be compatible with pyqtgraph.mkBrush()
        """
        self.region.setHoverBrush(brush)

    @staticmethod
    def _normalize_bounds(bounds: Union[None, Tuple[Union[None, float], Union[None, float]]]
                          ) -> Tuple[Union[None, float], Union[None, float]]:
        if bounds is None:
            bounds = (None, None)
        else:
            # Replace inf values by None
            try:
                if isinf(bounds[0]):
                    bounds = (None, bounds[1])
            except TypeError:
                pass
            try:
                if isinf(bounds[1]):
                    bounds = (bounds[0], None)
            except TypeError:
                pass
            # Sort spans in ascending order
            try:
                bounds = tuple(sorted(bounds))
            except TypeError:
                bounds = tuple(bounds)
        return bounds

    def _region_changed(self, obj: Optional[_LinearRegionItem] = None) -> None:
        if self._is_dragged:
            is_start = False
        else:
            self._is_dragged = True
            is_start = True
            if self.region.orientation == 'vertical':
                self._start_area = tuple(pos[0] for pos in self.region.startPositions)
            else:
                self._start_area = tuple(pos[1] for pos in self.region.startPositions)

        if self.region.moving:
            is_finished = False
        else:
            self._is_dragged = False
            is_finished = True

        self.sigAreaDragged.emit(self._start_area, self.area, is_start, is_finished)

    def _region_change_finished(self, obj: Optional[_LinearRegionItem] = None) -> None:
        current_area = self.area
        if self._is_dragged:
            self._is_dragged = False
            self.sigAreaDragged.emit(self._start_area, current_area, False, True)
        self.sigAreaChanged.emit(current_area)


class Rectangle(QtCore.QObject):
    """
    """

    _default_pen = {'color': '#00ff00', 'width': 1}
    _default_hover_pen = {'color': '#ffff00', 'width': 1}

    sigAreaChanged = QtCore.Signal(tuple)  # current_area
    # start_area, current_area, is_start, is_finished
    sigAreaDragged = QtCore.Signal(tuple, tuple, bool, bool)

    def __init__(self,
                 viewbox: ViewBox,
                 position: Optional[Tuple[float, float]] = (0, 0),
                 size: Optional[Tuple[float, float]] = (1, 1),
                 edge_handles: Optional[bool] = False,
                 corner_handles: Optional[bool] = False,
                 apply_bounds_to_center: Optional[bool] = False,
                 bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 movable: Optional[bool] = True,
                 resizable: Optional[bool] = True,
                 pen: Optional[Any] = None,
                 hover_pen: Optional[Any] = None
                 ) -> None:
        super().__init__(parent=viewbox)
        if position is None:
            position = (0, 0)
        if size is None:
            size = (1, 1)
        if movable is None:
            movable = True
        if resizable is None:
            resizable = True
        if pen is None:
            pen = self._default_pen
        if hover_pen is None:
            hover_pen = self._default_hover_pen

        self.roi = _RectangleROI(pos=position,
                                 size=size,
                                 apply_bounds_to_center=apply_bounds_to_center,
                                 bounds=bounds,
                                 movable=movable,
                                 resizable=resizable,
                                 pen=pen,
                                 hoverPen=hover_pen,
                                 handlePen=pen,
                                 handleHoverPen=hover_pen)

        self.roi.sigRegionChanged.connect(self._roi_changed)
        self.roi.sigRegionChangeFinished.connect(self._roi_change_finished)
        self.roi.sigRegionChangeStarted.connect(self._roi_change_started)

        self._z_value = None
        self._corner_handles = bool(corner_handles)
        self._edge_handles = bool(edge_handles)
        self._is_dragged = False
        self._start_area = self.area

        if resizable:
            self._add_handles()

        self.show()

    @property
    def movable(self) -> bool:
        return self.roi.translatable

    def set_movable(self, movable: bool) -> None:
        self.roi.translatable = bool(movable)

    @property
    def resizable(self) -> bool:
        return self.roi.resizable

    def set_resizable(self, resizable: bool) -> None:
        resizable = bool(resizable)
        changed = resizable != self.roi.resizable
        self.roi.resizable = resizable
        if changed:
            if resizable:
                self._add_handles()
            else:
                self._remove_handles()

    @property
    def z_value(self) -> Union[None, int]:
        return self._z_value

    def set_z_value(self, value: int) -> None:
        """ (Un-)Set the crosshair movable (draggable by mouse cursor) """
        self.roi.setZValue(value)
        self._z_value = value

    @property
    def position(self) -> Tuple[float, float]:
        return self.roi.area[0]

    def set_position(self, position: Tuple[float, float]) -> None:
        if not self._is_dragged:
            self.roi.set_area(position=position)

    @property
    def size(self) -> Tuple[float, float]:
        return self.roi.area[1]

    def set_size(self, size: Tuple[float, float]) -> None:
        if not self._is_dragged:
            self.roi.set_area(size=(abs(size[0]), abs(size[1])))

    @property
    def min_size(self) -> Tuple[float, float]:
        return self.roi.min_size

    def set_min_size(self, size: Tuple[float, float]) -> None:
        if not self._is_dragged:
            self.roi.set_min_size(size)

    @property
    def area(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        return self.roi.area

    def set_area(self,
                 position: Optional[Tuple[float, float]] = None,
                 size: Optional[Tuple[float, float]] = None
                 ) -> None:
        if not self._is_dragged:
            self.roi.set_area(position=position, size=size)

    @property
    def bounds(self) -> List[Tuple[Union[None, float], Union[None, float]]]:
        return self.roi.bounds

    def set_bounds(self,
                   bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
                   ) -> None:
        return self.roi.set_bounds(bounds)

    def show(self):
        view = self.parent()
        if self.roi not in view.addedItems:
            view.addItem(self.roi)
            if self._z_value is not None:
                self.roi.setZValue(self._z_value)

    def hide(self):
        view = self.parent()
        if self.roi in view.addedItems:
            view.removeItem(self.roi)

    def set_pen(self, pen: Any) -> None:
        """ Given parameter must be compatible with pyqtgraph.mkPen() """
        self.roi.setPen(pen)

    def set_hover_pen(self, pen: Any) -> None:
        """ Given parameter must be compatible with pyqtgraph.mkPen() """
        self.roi.setHoverPen(pen)

    def _add_handles(self) -> None:
        if self._corner_handles:
            self.roi.addScaleHandle([1, 1], [0, 0])
            self.roi.addScaleHandle([0, 0], [1, 1])
            self.roi.addScaleHandle([1, 0], [0, 1])
            self.roi.addScaleHandle([0, 1], [1, 0])
        if self._edge_handles:
            self.roi.addScaleHandle([1, 0.5], [0, 0.5])
            self.roi.addScaleHandle([0.5, 1], [0.5, 0])
            self.roi.addScaleHandle([0, 0.5], [1, 0.5])
            self.roi.addScaleHandle([0.5, 0], [0.5, 1])

    def _remove_handles(self) -> None:
        while len(self.roi.handles) > 0:
            self.roi.removeHandle(-1)

    def _roi_change_started(self, roi: Optional[_ROI] = None) -> None:
        self._is_dragged = True
        self._start_area = self.area
        self.sigAreaDragged.emit(self._start_area, self._start_area, True, False)

    def _roi_change_finished(self, roi: Optional[_ROI] = None) -> None:
        if self._is_dragged:
            self._is_dragged = False
            self.sigAreaDragged.emit(self._start_area, self.area, False, True)

    def _roi_changed(self, roi: Optional[_ROI] = None) -> None:
        current_area = self.area
        if self._is_dragged:
            self.sigAreaDragged.emit(self._start_area, current_area, False, False)
        self.sigAreaChanged.emit(current_area)


class InfiniteCrosshairRectangle(Rectangle):
    """ Represents a crosshair (two perpendicular infinite lines) and a finite sized rectangle ROI
    around the intersection
    """

    def __init__(self,
                 viewbox: ViewBox,
                 position: Optional[Tuple[float, float]] = (0, 0),
                 size: Optional[Tuple[float, float]] = (1, 1),
                 edge_handles: Optional[bool] = False,
                 corner_handles: Optional[bool] = False,
                 apply_bounds_to_center: Optional[bool] = True,
                 bounds: Optional[Sequence[Tuple[Union[None, float], Union[None, float]]]] = None,
                 movable: Optional[bool] = True,
                 resizable: Optional[bool] = True,
                 pen: Optional[Any] = None,
                 hover_pen: Optional[Any] = None
                 ) -> None:
        if pen is None:
            pen = self._default_pen
        if hover_pen is None:
            hover_pen = self._default_hover_pen
        bounds = _RectangleROI.normalize_bounds(bounds)
        self.vline = _InfiniteLine(pos=position[0],
                                   bounds=bounds[0],
                                   angle=90,
                                   movable=movable,
                                   pen=pen,
                                   hoverPen=hover_pen)
        self.hline = _InfiniteLine(pos=position[1],
                                   bounds=bounds[1],
                                   angle=0,
                                   movable=movable,
                                   pen=pen,
                                   hoverPen=hover_pen)

        super().__init__(viewbox=viewbox,
                         position=position,
                         size=size,
                         edge_handles=edge_handles,
                         corner_handles=corner_handles,
                         apply_bounds_to_center=apply_bounds_to_center,
                         bounds=bounds,
                         movable=movable,
                         resizable=resizable,
                         pen=pen,
                         hover_pen=hover_pen)

        self._line_is_dragged = False

        self.vline.sigDragged.connect(self._line_dragged)
        self.vline.sigPositionChanged.connect(self._line_changed)
        self.vline.sigPositionChangeFinished.connect(self._line_drag_finished)
        self.hline.sigDragged.connect(self._line_dragged)
        self.hline.sigPositionChanged.connect(self._line_changed)
        self.hline.sigPositionChangeFinished.connect(self._line_drag_finished)

    def set_movable(self, movable: bool) -> None:
        movable = bool(movable)
        self.roi.translatable = movable
        self.vline.setMovable(movable)
        self.hline.setMovable(movable)

    def set_z_value(self, value: int) -> None:
        """ (Un-)Set the crosshair movable (draggable by mouse cursor) """
        if self._z_value is not None:
            self.roi.setZValue(value)
            self.vline.setZValue(value)
            self.hline.setZValue(value)
            self._z_value = value

    def set_position(self, position: Tuple[float, float]) -> None:
        super().set_position(position)
        if not self._line_is_dragged:
            self._move_lines_to_roi()

    def set_size(self, size: Tuple[float, float]) -> None:
        super().set_size(size)
        self._move_lines_to_roi()

    def set_area(self,
                 position: Optional[Tuple[float, float]] = None,
                 size: Optional[Tuple[float, float]] = None
                 ) -> None:
        super().set_area(position=position, size=size)
        if not self._line_is_dragged:
            self._move_lines_to_roi()

    def set_bounds(self,
                   bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
                   ) -> None:
        super().set_bounds(bounds)
        x_bounds, y_bounds = self.bounds
        self.vline.setBounds(x_bounds)
        self.hline.setBounds(y_bounds)

    def show(self):
        view = self.parent()
        if self.hline not in view.addedItems:
            view.addItem(self.hline)
        if self.vline not in view.addedItems:
            view.addItem(self.vline)
        if self.roi not in view.addedItems:
            view.addItem(self.roi)
        self.set_z_value(self._z_value)

    def hide(self):
        view = self.parent()
        if self.roi in view.addedItems:
            view.removeItem(self.roi)
        if self.hline in view.addedItems:
            view.removeItem(self.hline)
        if self.vline in view.addedItems:
            view.removeItem(self.vline)

    def set_pen(self, pen: Any) -> None:
        """ Given parameter must be compatible with pyqtgraph.mkPen() """
        self.roi.setPen(pen)
        self.vline.setPen(pen)
        self.hline.setPen(pen)

    def set_hover_pen(self, pen: Any) -> None:
        """ Given parameter must be compatible with pyqtgraph.mkPen() """
        self.roi.setHoverPen(pen)
        self.vline.setHoverPen(pen)
        self.hline.setHoverPen(pen)

    def _add_handles(self) -> None:
        if self._corner_handles:
            self.roi.addScaleHandle([1, 1], [0.5, 0.5])
        if self._edge_handles:
            self.roi.addScaleHandle([1, 0.5], [0.5, 0.5])
            self.roi.addScaleHandle([0.5, 1], [0.5, 0.5])

    def _move_lines_to_roi(self) -> None:
        x_pos, y_pos = self.position
        self.vline.blockSignals(True)
        self.hline.blockSignals(True)
        try:
            self.vline.setValue(x_pos)
            self.hline.setValue(y_pos)
        finally:
            self.vline.blockSignals(False)
            self.hline.blockSignals(False)

    def _line_dragged(self, line: Optional[_InfiniteLine] = None) -> None:
        if self._line_is_dragged:
            is_start = False
        else:
            self._line_is_dragged = True
            self._start_area = self.area
            is_start = True

        new_pos = (self.vline.value(), self.hline.value())
        self.roi.blockSignals(True)
        try:
            self.roi.set_area(position=new_pos)
        finally:
            self.roi.blockSignals(False)
        current_area = self.roi.area

        self.sigAreaDragged.emit(self._start_area, current_area, is_start, False)

    def _line_drag_finished(self, line: Optional[_InfiniteLine] = None) -> None:
        if self._line_is_dragged:
            self._line_is_dragged = False
            self.sigAreaDragged.emit(self._start_area, self.area, False, True)

    def _line_changed(self, line: Optional[_InfiniteLine] = None) -> None:
        self.sigAreaChanged.emit(self.area)

    def _roi_changed(self, roi: Optional[_ROI] = None) -> None:
        self._move_lines_to_roi()
        super()._roi_changed(roi)



# -*- coding: utf-8 -*-
"""
This file contains Qudi methods for handling real-world values with units.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('create_formatted_output', 'get_relevant_digit', 'get_si_norm', 'get_unit_prefix_dict',
           'round_value_to_error', 'ScaledFloat')

import math
import numpy as np
try:
    import pyqtgraph.functions as fn
except ImportError:
    fn = None


def get_unit_prefix_dict():
    """Return the dictionary, which assigns the prefix of a unit to its proper order of magnitude.

    Parameters
    ----------
    None

    Returns
    -------
    dict 
        Keys are string prefix and values are magnitude values.
    """
    unit_prefix_dict = {
        'y': 1e-24,
        'z': 1e-21,
        'a': 1e-18,
        'f': 1e-15,
        'p': 1e-12,
        'n': 1e-9,
        'µ': 1e-6,
        'm': 1e-3,
        '': 1,
        'k': 1e3,
        'M': 1e6,
        'G': 1e9,
        'T': 1e12,
        'P': 1e15,
        'E': 1e18,
        'Z': 1e21,
        'Y': 1e24
    }
    return unit_prefix_dict


class ScaledFloat(float):
    """
    Format code 'r' for scaled output.

    Examples
    --------
    '{:.0r}A'.format(ScaledFloat(50))       --> 50 A
    '{:.1r}A'.format(ScaledFloat(1.5e3))    --> 1.5 kA
    '{:.1r}A'.format(ScaledFloat(2e-3))     --> 2.0 mA
    '{:rg}A'.format(ScaledFloat(2e-3))      --> 2 mA
    '{:rf}A'.format(ScaledFloat(2e-3))      --> 2.000000 mA
    """

    @property
    def scale(self):
        """
        Returns the scale. (No prefix if 0)

        Examples
        --------
        1e-3: m
        1e6: M
        """

        # Zero makes the log crash and should not have a prefix
        if self == 0:
            return ''

        exponent = math.floor(math.log10(abs(self)) / 3)
        if exponent < -8:
            exponent = -8
        if exponent > 8:
            exponent = 8
        prefix = 'yzafpnµm kMGTPEZY'
        return prefix[8 + exponent].strip()

    @property
    def scale_val(self):
        """Returns the scale value which can be used to devide the actual value.

        Examples
        --------
        m: 1e-3
        M: 1e6
        """
        scale_str = self.scale
        return get_unit_prefix_dict()[scale_str]

    def __format__(self, fmt):
        """
        Formats the string using format fmt.

        r for scaled output.

        Parameters
        ----------
        fmt : str 
            Format string.
        """
        autoscale = False
        if len(fmt) >= 2:
            if fmt[-2] == 'r':
                autoscale = True
                fmt = fmt[:-2] + fmt[-1]
            elif fmt[-1] == 'r':
                autoscale = True
                fmt = fmt[:-1] + 'f'
        elif fmt[-1] == 'r':
            autoscale = True
            fmt = fmt[:-1] + 'f'
        if autoscale:
            scale = self.scale
            if scale == 'u':
                index = 'micro'
            else:
                index = scale
            value = self / get_unit_prefix_dict()[index]
            return '{:s} {:s}'.format(value.__format__(fmt), scale)
        else:
            return super().__format__(fmt)


def create_formatted_output(param_dict, num_sig_digits=5):
    """Display a parameter set nicely in SI units.

    Parameters
    ----------
    param_dict : dict
        Dictionary with entries being dictionaries with two needed keywords 'value' and 'unit' and one
        optional keyword 'error'. Add the proper items to the specified keywords.
        Note that if no error is specified, no proper rounding (and therefore displaying) can be
        guaranteed.
    num_sig_digits : int, optional
        The number of significant digits will be taken if the rounding procedure was not
        successful at all. Default is 5.

    Returns
    -------
    str
        A nicely formatted string.
        
    Notes
    -----
    The absolute tolerance to a zero is set to 1e-18.
        
    Examples
    --------

    Example of a param dict:
    
    param_dict = {'Rabi frequency': {'value': 123.43, 'error': 0.321, 'unit': 'Hz'},
                  'ODMR contrast':  {'value': 2.563423, 'error': 0.523, 'unit': '%'},
                  'Fidelity':       {'value': 0.783, 'error': 0.2222, 'unit': ''}}
    
    If you want to access the value of the Fidelity, then you can do that via:
    
    >>> param_dict['Fidelity']['value']
    
    or on the error of the ODMR contrast:
    
    >>> param_dict['ODMR contrast']['error']
    """
    if fn is None:
        raise RuntimeError('Function "create_formatted_output" requires pyqtgraph.')

    output_str = ''
    atol = 1e-18    # absolute tolerance for the detection of zero.

    for entry in param_dict:
        if param_dict[entry].get('error') is not None:

            value, error, digit = round_value_to_error(
                param_dict[entry]['value'], param_dict[entry]['error'])

            if (np.isclose(value, 0.0, atol=atol)
                    or np.isnan(error)
                    or np.isclose(error, 0.0, atol=atol)
                    or np.isinf(error)):
                sc_fact, unit_prefix = fn.siScale(param_dict[entry]['value'])
                str_val = '{0:.{1}e}'.format(
                    param_dict[entry]['value'], num_sig_digits - 1)
                if np.isnan(float(str_val)):
                    value = np.NAN
                elif np.isinf(float(str_val)):
                    value = np.inf
                else:
                    value = float('{0:.{1}e}'.format(
                        param_dict[entry]['value'], num_sig_digits - 1))

            else:
                # the factor 10 moves the displayed digit by one to the right,
                # so that the values from 100 to 0.1 are displayed within one
                # range, rather then from the value 1000 to 1, which is
                # default.
                sc_fact, unit_prefix = fn.siScale(error * 10)
            output_str += '{0}: ({1} \u00B1 {2}) {3}{4} \n'.format(entry, round(value * sc_fact,
                                                                                num_sig_digits - 1),
                                                                   round(error * sc_fact,
                                                                         num_sig_digits - 1),
                                                                   unit_prefix,
                                                                   param_dict[entry]['unit'])
        else:
            output_str += '{0}: '.format(entry) + fn.siFormat(param_dict[entry]['value'],
                                                              precision=num_sig_digits,
                                                              suffix=param_dict[entry][
                                                                  'unit']) + ' (fixed) \n'
    return output_str


def round_value_to_error(value, error):
    """The scientifically correct way of rounding a value according to an error.

    Parameters
    ----------
    value : float or int 
        The measurement value.
    error : float or int
        The error for that measurement value.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        
        float
            The rounded value according to the error.
        float
            The rounded error.
        int
            The digit to which the rounding procedure was performed. A positive
            number indicates the position of the digit right from the comma, zero means
            the first digit left from the comma, and negative numbers are the digits left
            from the comma. This follows the convention used in the native `round` method
            and `numpy.round`.

    Notes
    -----
    - The input type of `value` or `error` will not be changed. If `float` is the input, `float` will be the output; the same applies to `integer`.
    - This method does not return strings, as each display method might want to display the rounded values in a different way (in exponential representation, in a different magnitude, etc.).
    - This function can handle an invalid error, i.e., if the error is zero, NaN, or infinite. The absolute tolerance to detect a number as zero is set to 1e-18.

    Procedure explanation:
    The scientific way of displaying a measurement result in the presence of an error is applied here. It follows this procedure:
    Take the first leading non-zero number in the error value and check whether the number is a digit within 3 to 9. If so, the rounding value
    is the specified digit. Otherwise, if the first leading digit is 1 or 2, then the next right digit is the rounding value.
    The error is rounded according to that digit, and the same applies to the value.

    Examples
    --------

    Example 1:

    >>> x_meas = 2.05650234
    >>> delta_x = 0.0634
    >>> result = some_function(x_meas, delta_x)
    >>> print(result)
    (2.06, 0.06, 2)

    Example 2:

    >>> x_meas = 0.34545
    >>> delta_x = 0.19145
    >>> result = some_function(x_meas, delta_x)
    >>> print(result)
    (0.35, 0.19, 2)

    Example 3:

    >>> x_meas = 239579.23
    >>> delta_x = 1289.234
    >>> result = some_function(x_meas, delta_x)
    >>> print(result)
    (239600.0, 1300.0, -2)

    Example 4:

    >>> x_meas = 961453
    >>> delta_x = 3789
    >>> result = some_function(x_meas, delta_x)
    >>> print(result)
    (961000, 4000, -3)
    """

    atol = 1e-18    # absolute tolerance for the detection of zero.

    # check if error is zero, since that is an invalid input!
    if np.isclose(error, 0.0, atol=atol) or np.isnan(error) or np.isinf(error):
        # self.log.error('Cannot round to the error, since either a zero error ')
        # logger.warning('Cannot round to the error, since either a zero error '
        #            'value was passed for the number {0}, or the error is '
        #            'NaN: Error value: {1}. '.format(value, error))

        # set the round digit to float precision
        round_digit = -12

        return value, error, round_digit

    # error can only be positive!
    log_val = np.log10(abs(error))

    if log_val < 0:
        round_digit = -(int(log_val) - 1)
    else:
        round_digit = -(int(log_val))

    first_err_digit = '{:e}'.format(error)[0]

    if first_err_digit in ('1', '2'):
        round_digit += 1

    # Use the python round function, since np.round uses the __repr__ conversion
    # function which shows enough digits to unambiguously identify the number.
    # But the __str__ conversion should round the number to a reasonable number
    # of digits, which is the standard output of the python round function.
    # Therefore take the python round function.

    return round(value, round_digit), round(error, round_digit), round_digit


def get_relevant_digit(entry):
    """By using log10, abs and int operations, the proper relevant digit is
    obtained.

    Parameters
    ----------
    entry : float
        
    Returns
    -------
    int
        The leading relevant exponent.
    """
    # the log10 can only be calculated of a positive number.
    entry = np.abs(entry)

    # the log of zero crashes, so return 0
    if entry == 0:
        return 0

    if np.log10(entry) >= 0:
        return int(np.log10(entry))
    else:
        # catch the asymmetric behaviour of the log and int operation.
        return int(int(np.abs(np.log10(entry))) + 1 + np.log10(entry)) - (
                    int(np.abs(np.log10(entry))) + 1)


def get_si_norm(entry):
    """A rather different way to display the value in SI notation.

    Parameters
    ----------
    entry : float 
        The float number from which normalization factor should
        be obtained.

    Returns
    -------
    norm_val : float
        The value in a normalized representation.
    normalization : float
        The factor by which to divide the number.
    """
    val = get_relevant_digit(entry)
    fact = int(val / 3)
    power = int(3 * fact)
    norm = 10 ** power

    return entry / norm, norm



# -*- coding: utf-8 -*-
"""
This file contains the Qudi thread manager singleton class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import weakref
from functools import partial
from PySide2 import QtCore

from qudi.util.mutex import RecursiveMutex
from qudi.core.logger import get_logger

logger = get_logger(__name__)


class ThreadManager(QtCore.QAbstractListModel):
    """This class keeps track of all the QThreads that are needed somewhere.

    Using this class is thread-safe.
    """
    _instance = None
    _lock = RecursiveMutex()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None or cls._instance() is None:
                obj = super().__new__(cls, *args, **kwargs)
                cls._instance = weakref.ref(obj)
                return obj
            raise RuntimeError(
                'Only one ThreadManager instance per process possible (Singleton). Please use '
                'ThreadManager.instance() to get a reference to the already created instance.'
            )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._threads = list()
        self._thread_names = list()

    @classmethod
    def instance(cls):
        with cls._lock:
            if cls._instance is None:
                return None
            return cls._instance()

    @property
    def thread_names(self):
        with self._lock:
            return self._thread_names.copy()

    def get_new_thread(self, name):
        """Create and return a new QThread with objectName <name>.

        Parameters
        ----------
        name : str
            Unique name of the thread.

        Returns
        -------
        QThread
            New thread, or None if creation failed.
        """
        with self._lock:
            logger.debug('Creating thread: "{0}".'.format(name))
            if name in self._thread_names:
                return None
            thread = QtCore.QThread()
            thread.setObjectName(name)
            self.register_thread(thread)
            return thread

    @QtCore.Slot(QtCore.QThread)
    def register_thread(self, thread):
        """Add QThread to ThreadManager.

        Parameters
        ----------
        thread : QtCore.QThread
            Thread to register with a unique objectName.
        """
        with self._lock:
            name = thread.objectName()
            if name in self._thread_names:
                if self.get_thread_by_name(name) is thread:
                    return None
                raise RuntimeError(
                    f'Different thread with name "{name}" already registered in ThreadManager'
                )

            row = len(self._threads)
            self.beginInsertRows(QtCore.QModelIndex(), row, row)
            self._threads.append(thread)
            self._thread_names.append(name)
            thread.finished.connect(
                partial(self.unregister_thread, name=name), QtCore.Qt.QueuedConnection)
            self.endInsertRows()

    @QtCore.Slot(object)
    def unregister_thread(self, name):
        """Remove thread from ThreadManager.

        Parameters
        ----------
        name : str
            Unique thread name.
        """
        with self._lock:
            if isinstance(name, QtCore.QThread):
                name = name.objectName()
            if name in self._thread_names:
                index = self._thread_names.index(name)
                if self._threads[index].isRunning():
                    self.quit_thread(name)
                    return
                logger.debug('Cleaning up thread {0}.'.format(name))
                self.beginRemoveRows(QtCore.QModelIndex(), index, index)
                del self._threads[index]
                del self._thread_names[index]
                self.endRemoveRows()

    @QtCore.Slot(object)
    def quit_thread(self, name):
        """Stop event loop of QThread.

        Parameters
        ----------
        name : str
            Unique thread name.
        """
        with self._lock:
            if isinstance(name, QtCore.QThread):
                thread = name
            else:
                thread = self.get_thread_by_name(name)
            if thread is None:
                logger.debug('You tried quitting a nonexistent thread {0}.'.format(name))
            else:
                logger.debug('Quitting thread {0}.'.format(name))
                thread.quit()

    @QtCore.Slot(object, int)
    def join_thread(self, name, time=None):
        """Wait for stop of QThread event loop.

        Parameters
        ----------
        name : str
            Unique thread name.
        time : int
            Timeout for waiting in milliseconds.
        """
        with self._lock:
            if isinstance(name, QtCore.QThread):
                thread = name
            else:
                thread = self.get_thread_by_name(name)
            if thread is None:
                logger.debug('You tried waiting for a nonexistent thread {0}.'.format(name))
            else:
                logger.debug('Waiting for thread {0} to end.'.format(name))
                if time is None:
                    thread.wait()
                else:
                    thread.wait(time)

    @QtCore.Slot(int)
    def quit_all_threads(self, thread_timeout=10000):
        """Stop event loop of all QThreads.
        """
        with self._lock:
            logger.debug('Quit all threads.')
            for thread in self._threads:
                thread.quit()
                if not thread.wait(int(thread_timeout)):
                    logger.error('Waiting for thread {0} timed out.'.format(thread.objectName()))

    def get_thread_by_name(self, name):
        """Get registered QThread instance by its objectName.

        Parameters
        ----------
        name : str
            Object name of the QThread to return.

        Returns
        -------
        QThread
            The registered thread object.
        """
        with self._lock:
            try:
                index = self._thread_names.index(name)
                return self._threads[index]
            except ValueError:
                return None

    # QAbstractListModel interface methods follow below
    def rowCount(self, parent=None, *args, **kwargs):
        """Gives the number of threads registered.

        Returns
        -------
        int
            Number of threads.
        """
        with self._lock:
            return len(self._threads)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """Data for the list view header.

        Parameters
        ----------
        section : int
            Column/row index to get header data for.
        orientation : QtCore.Qt.Orientation
            Orientation of header (horizontal or vertical).
        role : QtCore.ItemDataRole
            Data access role.

        Returns
        -------
        str
            Header data for the given column/row and role.
        """
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal and section == 0:
            return 'Thread Name'
        return None

    def data(self, index, role):
        """Get data from model for a given cell. Data can have a role that affects display.

        Parameters
        ----------
        index : QtCore.QModelIndex
            Cell for which data is requested.
        role : QtCore.Qt.ItemDataRole
            Data access role of the request.

        Returns
        -------
        QVariant
            Data for the given cell and role.
        """
        with self._lock:
            row = index.row()
            if index.isValid() and role == QtCore.Qt.DisplayRole and 0 <= row < len(self._threads):
                if index.column() == 0:
                    return self._thread_names[row]
            return None

    def flags(self, index):
        """Determines what can be done with entry cells in the table view.

        Parameters
        ----------
        index : QModelIndex
            Cell for which the flags are requested.

        Returns
        -------
        Qt.ItemFlags
            Actions allowed for this cell.
        """
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable



# -*- coding: utf-8 -*-
"""
This file contains the Qudi error dialog class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import traceback
from datetime import datetime
from collections import deque
from qudi.util.mutex import RecursiveMutex
from PySide2 import QtWidgets, QtCore


class ErrorDialog(QtWidgets.QDialog):
    """This class provides a popup window for notification with the option to
      show the next error popup in the queue and to show the log window where
      you can see the traceback for an exception.
    """

    _stylesheet_map = {'error'   : 'font-weight: bold; color: #F11000;',
                       'critical': 'font-weight: bold; color: #FF00FF;'}

    def __init__(self, *args, **kwargs):
        """Create an ErrorDialog object.
        """
        super().__init__(*args, **kwargs)

        self._thread_lock = RecursiveMutex()
        self._error_queue = deque()  # queued individual error messages to display

        # Set up dialog window
        self.setWindowTitle('Qudi Error')
        self.setWindowFlags((QtCore.Qt.Dialog |
                             QtCore.Qt.CustomizeWindowHint |
                             QtCore.Qt.WindowSystemMenuHint |
                             QtCore.Qt.WindowTitleHint) & (~QtCore.Qt.WindowCloseButtonHint))
        self.setModal(True)
        screen_size = QtWidgets.QApplication.instance().primaryScreen().availableSize()
        screen_width = screen_size.width()
        screen_height = screen_size.height()
        self._default_size = ((screen_width * 3) // 8,
                              (screen_height * 3) // 8,
                              screen_width // 4,
                              screen_height // 4)
        self.setGeometry(*self._default_size)
        self.setMinimumSize(screen_width // 6, screen_height // 8)

        # Set up header label widget
        self.header_label = QtWidgets.QLabel()
        self.header_label.setFocusPolicy(QtCore.Qt.NoFocus)
        self.header_label.setTextInteractionFlags(
            QtCore.Qt.TextSelectableByMouse | QtCore.Qt.LinksAccessibleByMouse
        )
        self.header_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                     QtWidgets.QSizePolicy.Preferred)

        # Set up scrollable message label widget
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFocusPolicy(QtCore.Qt.NoFocus)
        self.msg_label = QtWidgets.QLabel()
        self.msg_label.setFocusPolicy(QtCore.Qt.NoFocus)
        self.msg_label.setTextInteractionFlags(
            QtCore.Qt.TextSelectableByMouse | QtCore.Qt.LinksAccessibleByMouse
        )
        self.msg_label.setWordWrap(True)
        self.msg_label.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        self.msg_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                     QtWidgets.QSizePolicy.Expanding)
        scroll_area.setWidget(self.msg_label)

        # Set up disable checkbox
        self.disable_checkbox = QtWidgets.QCheckBox('Disable error message popups')

        # Set up buttons and group them in a layout
        self.dismiss_button = QtWidgets.QPushButton('Dismiss')
        self.next_button = QtWidgets.QPushButton('Show next error')
        btn_layout = QtWidgets.QHBoxLayout()
        btn_layout.addStretch()
        btn_layout.addWidget(self.next_button)
        btn_layout.addWidget(self.dismiss_button)

        # Set up dialog main layout and add all widgets to it
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(3, 3, 3, 3)
        main_layout.addWidget(self.header_label)
        main_layout.addWidget(scroll_area)
        main_layout.addWidget(self.disable_checkbox)
        main_layout.addLayout(btn_layout)
        self.setLayout(main_layout)

        # Connect button click signals to slots
        self.dismiss_button.clicked.connect(self.accept)
        self.next_button.clicked.connect(self.show_next_error)
        self.accepted.connect(self._accepted_callback)

    @QtCore.Slot(object)
    def new_error(self, data):
        """
        Show a new error log entry.

        Parameters
        ----------
        data : logging.LogRecord
            Log record as returned from the logging module.
        """
        with self._thread_lock:
            self._error_queue.append(data)
            if self.isVisible():
                self._update_next_button()
            else:
                self.setGeometry(*self._default_size)
                self.show_next_error()

    @QtCore.Slot()
    def show_next_error(self):
        with self._thread_lock:
            if len(self._error_queue) > 0:
                err = self._error_queue.popleft()
                self._update_next_button()
                time_str = datetime.fromtimestamp(err.created).strftime('%Y-%m-%d %H:%M:%S')
                message = err.message if hasattr(err, 'message') else err.msg
                if err.exc_info is not None:
                    message += '\n\n{0}'.format(traceback.format_exception(*err.exc_info)[-1][:-1])
                    tb = '\n'.join(traceback.format_exception(*err.exc_info)[:-1])
                    if tb:
                        message += '\n{0}'.format(tb)

                self.header_label.setStyleSheet(self._stylesheet_map[err.levelname])
                self.header_label.setText('Error in {0} ({1}):'.format(err.name, time_str))
                self.msg_label.setText(message)
                if self.enabled:
                    self.show()
                    self.activateWindow()

    @property
    def enabled(self):
        """Property holding the enabled flag for this error message popup.

        Returns
        -------
        bool
            Flag indicating whether error message popups are enabled (`True`) or disabled (`False`).
        """
        with self._thread_lock:
            return not self.disable_checkbox.isChecked()

    @QtCore.Slot(bool)
    def set_enabled(self, enable):
        with self._thread_lock:
            self.disable_checkbox.setChecked(not bool(enable))

    @QtCore.Slot()
    def _accepted_callback(self):
        with self._thread_lock:
            self._error_queue.clear()

    def _update_next_button(self):
        with self._thread_lock:
            msg_number = len(self._error_queue)
            btn_enabled = self.next_button.isEnabled()
            self.next_button.setText('Show next error ({0:d} more)'.format(msg_number))
            if msg_number == 0 and btn_enabled:
                self.next_button.setEnabled(False)
                self.dismiss_button.setFocus()
            elif msg_number > 0 and not btn_enabled:
                self.next_button.setEnabled(True)
                self.next_button.setFocus()

    @QtCore.Slot()
    def reject(self):
        """Override reject slot in order to prevent rejection of this QDialog.
        """
        if not self.dismiss_button.hasFocus():
            self.dismiss_button.setFocus()

